import concurrent.futures
import csv
from enum import Enum
import json
import logging
import math
import sys
import traceback
import typing

import synapseclient
from synapseclient.core.constants import concrete_types
from synapseclient.core import pool_provider
from synapseclient.core import utils
from synapseclient.table import join_column_names
from synapseclient.core.upload.multipart_upload import (
    MAX_NUMBER_OF_PARTS,
    multipart_copy,
    shared_executor,
)

"""
Contains functions for migrating the storage location of Synapse entities.
Entities can be updated or moved so that their underlying file handles are stored
in the new location.
"""


def test_import_sqlite3():
    # sqlite3 is part of the Python standard library and is available on the vast majority
    # of Python installations and doesn't require any additional software on the system.
    # it may be unavailable in some rare cases though (for example Python compiled from source
    # without ay sqlite headers available). we dynamically import it when used to avoid making
    # this dependency hard for all client usage, however.
    try:
        import sqlite3  # noqa
    except ImportError:
        sys.stderr.write("""\nThis operation requires the sqlite3 module which is not available on this
installation of python. Using a Python installed from a binary package or compiled from source with sqlite
development headers available should ensure that the sqlite3 module is available.""")
        raise


# we use a much larger default part size for part copies than we would for part uploads.
# with part copies the data transfer is within AWS so don't need to concern ourselves
# with upload failures of the actual bytes.
# this value aligns with what some AWS client libraries use e.g.
# https://github.com/aws/aws-sdk-java/blob/1.11.995/aws-java-sdk-s3/src/main/java/com/amazonaws/services/s3/transfer/TransferManagerConfiguration.java#L46
DEFAULT_PART_SIZE = 100 * utils.MB


class _MigrationStatus(Enum):
    # an internal enum for use within the sqlite db
    # to track the state of entities as they are indexed
    # and then migrated.
    INDEXED = 1
    MIGRATED = 2
    ALREADY_MIGRATED = 3
    ERRORED = 4


class _MigrationType(Enum):
    # container types (projects and folders) are only used during the indexing phase.
    # we record the containers we've indexed so we don't reindex them on a subsequent
    # run using the same db file (or reindex them after an indexing dry run)
    PROJECT = 1
    FOLDER = 2

    # files and table attached files represent file handles that are actually migrated
    FILE = 3
    TABLE_ATTACHED_FILE = 4

    @classmethod
    def from_concrete_type(cls, concrete_type):
        if concrete_type == concrete_types.PROJECT_ENTITY:
            return cls.PROJECT
        elif concrete_type == concrete_types.FOLDER_ENTITY:
            return cls.FOLDER
        elif concrete_type == concrete_types.FILE_ENTITY:
            return cls.FILE
        elif concrete_type == concrete_types.TABLE_ENTITY:
            return cls.TABLE_ATTACHED_FILE

        raise ValueError("Unhandled type {}".format(concrete_type))


class _MigrationKey(typing.NamedTuple):
    id: str
    type: _MigrationType
    version: int
    row_id: int
    col_id: int


def _get_row_dict(cursor, row, include_empty):
    return {
        col[0]: row[i] for i, col in enumerate(cursor.description)
        if (include_empty or row[i] is not None) and col[0] != 'rowid'
    }


class MigrationResult:
    """A MigrationResult is a proxy object to the underlying sqlite db.
    It provides a programmatic interface that allows the caller to iterate over the
    file handles that were migrated without having to connect to or know the schema
    of the sqlite db, and also avoids the potential memory liability of putting
    everything into an in memory data structure that could be a liability when
    migrating a huge project of hundreds of thousands/millions of entities.

    As this proxy object is not thread safe since it accesses an underlying sqlite db.
    """

    def __init__(self, syn, db_path):
        self._syn = syn
        self.db_path = db_path

    def get_counts_by_status(self):
        """
            Returns a dictionary of counts by the migration status of each indexed file/version.
            Keys are as follows:

                * INDEXED - the file/version has been indexed and will be migrated on a call to migrate_indexed_files
                * MIGRATED - the file/version has been migrated
                * ALREADY_MIGRATED - the file/version was already stored at the target storage location and no migration is needed
                * ERRORED - an error occurred while indexing or migrating the file/version
        """  # noqa
        import sqlite3
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()

            # for the purposes of these counts, containers (Projects and Folders) do not count.
            # we are counting actual files only
            result = cursor.execute(
                'select status, count(*) from migrations where type in (?, ?) group by status',
                (_MigrationType.FILE.value, _MigrationType.TABLE_ATTACHED_FILE.value)
            )

            counts_by_status = {status.name: 0 for status in _MigrationStatus}
            for row in result:
                status = row[0]
                count = row[1]
                counts_by_status[_MigrationStatus(status).name] = count

            return counts_by_status

    def get_migrations(self):
        """
        A generator yielding each file/version in the migration index.
        A dictionary of the properties of the migration row is yielded as follows:
            id - the Synapse id
            type - the concrete type of the entity
            version - the verson of the file entity (if applicable)
            row_id - the row of the table attached file (if applicable)
            col_id - the column id of the table attached file (if applicable)
            from_storage_location_id - the previous storage location id where the file/version was stored
            from_file_handle_id - the id file handle of the existing file/version
            to_file_handle_id - if migrated, the new file handle id
            status - one of INDEXED, MIGRATED, ALREADY_MIGRATED, ERRORED indicating the status of the file/version
            exception - if an error was encountered indexing/migrating the file/version its stack is here
        """
        import sqlite3
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()

            last_id = None
            column_names = None

            rowid = -1
            while True:
                results = cursor.execute(
                    """
                        select
                            rowid,

                            id,
                            type,
                            version,
                            row_id,
                            col_id,
                            from_storage_location_id,
                            from_file_handle_id,
                            to_file_handle_id,
                            file_size,
                            status,
                            exception
                        from migrations
                        where
                            rowid > ?
                            and type in (?, ?)
                        order by
                            rowid
                        limit ?
                    """,
                    (
                        rowid,
                        _MigrationType.FILE.value, _MigrationType.TABLE_ATTACHED_FILE.value,
                        _get_batch_size()
                    )
                )

                row_count = 0
                for row in results:
                    row_count += 1

                    # using the internal sqlite rowid for ordering only
                    rowid = row[0]

                    # exclude the sqlite internal rowid
                    row_dict = _get_row_dict(cursor, row, False)
                    entity_id = row_dict['id']
                    if entity_id != last_id:
                        # if the next row is dealing with a different entity than the last table
                        # id then we discard any cached column names we looked up
                        column_names = {}

                    row_dict['type'] = 'file' if row_dict['type'] == _MigrationType.FILE.value else 'table'

                    for int_arg in (
                            'version',
                            'row_id',
                            'from_storage_location_id',
                            'from_file_handle_id',
                            'to_file_handle_id'
                    ):
                        int_val = row_dict.get(int_arg)
                        if int_val is not None:
                            row_dict[int_arg] = int(int_val)

                    col_id = row_dict.pop('col_id', None)
                    if col_id is not None:
                        column_name = column_names.get(col_id)

                        # for usability we look up the actual column name from the id,
                        # but that involves a lookup so we cache them for re-use across
                        # rows that deal with the same table entity
                        if column_name is None:
                            column = self._syn.restGET("/column/{}".format(col_id))
                            column_name = column_names[col_id] = column['name']

                        row_dict['col_name'] = column_name

                    row_dict['status'] = _MigrationStatus(row_dict['status']).name

                    yield row_dict

                    last_id = entity_id

                if row_count == 0:
                    # out of rows
                    break

    def as_csv(self, path):
        """
        Output a flat csv file of the contents of the Migration index.
        Its columns are as follows:
            id - the Synapse id
            type - the concrete type of the entity
            version - the verson of the file entity (if applicable)
            row_id - the row of the table attached file (if applicable)
            col_name - the column name of the column the table attached file resides in (if applicable)
            from_storage_location_id - the previous storage location id where the file/version was stored
            from_file_handle_id - the id file handle of the existing file/version
            to_file_handle_id - if migrated, the new file handle id
            status - one of INDEXED, MIGRATED, ALREADY_MIGRATED, ERRORED indicating the status of the file/version
            exception - if an error was encountered indexing/migrating the file/version its stack is here

        """

        with open(path, 'w', newline='') as csv_file:
            csv_writer = csv.writer(csv_file)

            # headers
            csv_writer.writerow([
                'id',
                'type',
                'version',
                'row_id',
                'col_name',
                'from_storage_location_id',
                'from_file_handle_id',
                'to_file_handle_id',
                'status',
                'exception'
            ])

            for row_dict in self.get_migrations():
                row_data = [
                    row_dict['id'],
                    row_dict['type'],
                    row_dict.get('version'),
                    row_dict.get('row_id'),
                    row_dict.get('col_name'),
                    row_dict.get('from_storage_location_id'),
                    row_dict.get('from_file_handle_id'),
                    row_dict.get('to_file_handle_id'),
                    row_dict['status'],
                    row_dict.get('exception')
                ]

                csv_writer.writerow(row_data)


def _get_executor(syn):
    executor = pool_provider.get_executor(thread_count=syn.max_threads)

    # default the number of concurrent file copies to half the number of threads in the pool.
    # since we share the same thread pool between managing entity copies and the multipart
    # upload, we have to prevent thread starvation if all threads are consumed by the entity
    # code leaving none for the multipart copies
    max_concurrent_file_copies = max(int(syn.max_threads / 2), 1)
    return executor, max_concurrent_file_copies


def _get_batch_size():
    # just a limit on certain operations to put an upper bound on various
    # batch operations so they are chunked. a function to make it easily mocked.
    # don't anticipate needing to adjust this for any real activity
    return 500


def _ensure_schema(cursor):
    # ensure we have the sqlite schema we need to be able to record and sort our
    # entity file handle migration.

    # one-row table of a json dictionary records the parameters used to create the index
    cursor.execute("create table if not exists migration_settings (settings text not null)")

    # our representation of migratable file handles is flat including both file entities
    # and table attached files, so not all columns are applicable to both. row id and col id
    # are only used by table attached files, for example.
    cursor.execute(
        """
            create table if not exists migrations (
                id text not null,
                type integer not null,
                version integer null,
                row_id integer null,
                col_id integer null,

                parent_id null,
                status integer not null,
                exception text null,

                from_storage_location_id null,
                from_file_handle_id text null,
                to_file_handle_id text null,
                file_size integer null,

                primary key (id, type, row_id, col_id, version)
            )
        """
    )

    # we get counts grouping on status
    cursor.execute("create index if not exists ix_status on migrations(status)")

    # we check to see if there is already a migrated copy of a file handle before doing a copy
    cursor.execute(
        "create index if not exists ix_file_handle_ids on migrations(from_file_handle_id, to_file_handle_id)"
    )


def _wait_futures(conn, cursor, futures, pending_keys, return_when, continue_on_error):
    completed, futures = concurrent.futures.wait(futures, return_when=return_when)
    completed_file_handle_ids = set()

    for completed_future in completed:

        to_file_handle_id = None
        ex = None
        try:
            key, from_file_handle_id, to_file_handle_id = completed_future.result()
            completed_file_handle_ids.add(from_file_handle_id)
            status = _MigrationStatus.MIGRATED.value

        except _MigrationError as migration_ex:
            # for the purposes of recording and re-raise we're not interested in
            # the _MigrationError, just the underlying cause

            ex = migration_ex.__cause__
            key = migration_ex.key
            completed_file_handle_ids.add(migration_ex.from_file_handle_id)
            status = _MigrationStatus.ERRORED.value

        tb_str = ''.join(traceback.format_exception(type(ex), ex, ex.__traceback__)) if ex else None
        update_statement = """
            update migrations set
                status = ?,
                to_file_handle_id = ?,
                exception = ?
            where
                id = ?
                and type = ?
        """

        update_args = [status, to_file_handle_id, tb_str, key.id, key.type]
        for arg in ('version', 'row_id', 'col_id'):
            arg_value = getattr(key, arg)
            if arg_value is not None:
                update_statement += "and {} = ?\n".format(arg)
                update_args.append(arg_value)
            else:
                update_statement += "and {} is null\n".format(arg)

        cursor.execute(update_statement, tuple(update_args))
        conn.commit()

        pending_keys.remove(key)
        if not continue_on_error and ex:
            raise ex from None

    return futures, completed_file_handle_ids


def index_files_for_migration(
    syn: synapseclient.Synapse,
    entity,
    dest_storage_location_id: str,
    db_path: str,
    source_storage_location_ids: typing.Iterable[str] = None,
    file_version_strategy='new',
    include_table_files=False,
    continue_on_error=False,
):
    """
    Index the given entity for migration to a new storage location. This is the first step in migrating an entity
    to a new storage location using synapseutils.

    This function will create a sqlite database at the given db_path that can be subsequently passed
    to the migrate_indexed_files function for actual migration. This function itself does not modify the given entity
    in any way.

    :param syn:                         A Synapse object with user's login, e.g. syn = synapseclient.login()
    :param entity:                      A Synapse entity whose files should be migrated. Can be a Project, Folder,
                                        File entity, or Table entity. If it is a container (a Project or Folder) its
                                        contents will be recursively indexed.
    :param dest_storage_location_id:    The id of the new storage location to be migrated to.
    :param db_path:                     A path on disk where a sqlite db can be created to store the contents of the
                                        created index.
    :param source_storage_location_ids: An optional iterable of storage location ids that will be migrated. If provided,
                                        files outside of one of the listed storage locations will not be indexed
                                        for migration. If not provided, then all files not already in the destination
                                        storage location will be indexed for migrated. 
    :param file_version_strategy:   One of "new" (default), "all", "latest", "skip" as follows:

                                        * "new" - will create a new version of file entities in the new storage location, leaving existing versions unchanged
                                        * "all" - all existing versions will be migrated in place to the new storage location
                                        * "latest" - the latest version will be migrated in place to the new storage location
                                        * "skip" - skip migrating file entities. use this e.g. if wanting to e.g. migrate table attached files in a container while leaving the files unchanged
    :param include_table_files:     Whether to migrate files attached to tables. If False (default) then e.g. only
                                        file entities in the container will be migrated and tables will be untouched.
    :param continue_on_error:       Whether any errors encountered while indexing an entity (access etc) will be raised
                                        or instead just recorded in the index while allowing the index creation
                                        to continue. Default is False (any errors are raised).

    :return:                        A MigrationResult object that can be used to inspect the contents of the index
                                        or output the index to a CSV for manual inspection.
    """  # noqa
    root_id = utils.id_of(entity)

    # accept an Iterable, but easier to work internally if we can assume a list of strings
    source_storage_location_ids = [str(s) for s in source_storage_location_ids or []]

    file_version_strategies = {'new', 'all', 'latest', 'skip'}
    if file_version_strategy not in file_version_strategies:
        raise ValueError(
            "Invalid file_version_strategy: {}, must be one of {}".format(
                file_version_strategy,
                file_version_strategies
            )
        )

    if file_version_strategy == 'skip' and not include_table_files:
        raise ValueError('Skipping both files entities and table attached files, nothing to migrate')

    _verify_storage_location_ownership(syn, dest_storage_location_id)

    test_import_sqlite3()
    import sqlite3
    with sqlite3.connect(db_path) as conn:

        cursor = conn.cursor()
        _ensure_schema(cursor)

        _verify_index_settings(
            cursor,
            db_path,
            root_id,
            dest_storage_location_id,
            source_storage_location_ids,
            file_version_strategy,
            include_table_files
        )
        conn.commit()

        entity = syn.get(root_id, downloadFile=False)
        try:
            _index_entity(
                conn,
                cursor,
                syn,
                entity,
                None,
                dest_storage_location_id,
                source_storage_location_ids,
                file_version_strategy,
                include_table_files,
                continue_on_error,
            )

        except _IndexingError as indexing_ex:
            logging.exception(
                "Aborted due to failure to index entity %s of type %s. Use the continue_on_error option to skip "
                "over entities due to individual failures.",
                indexing_ex.entity_id,
                indexing_ex.concrete_type,
            )

            raise indexing_ex.__cause__

    return MigrationResult(syn, db_path)


def _confirm_migration(cursor, force, storage_location_id):
    # we proceed with migration if either using the force option or if
    # we can prompt the user with the count if items that are going to
    # be migrated and receive their confirmation from shell input
    confirmed = force
    if not force:
        count = cursor.execute(
            "select count(*) from migrations where status = ?",
            (_MigrationStatus.INDEXED.value,)
        ).fetchone()[0]

        if count == 0:
            logging.info("No items for migration.")

        elif sys.stdout.isatty():
            uinput = input("{} items for migration to {}. Proceed? (y/n)? ".format(
                count,
                storage_location_id
            ))
            confirmed = uinput.strip().lower() == 'y'

        else:
            logging.info(
                "%s items for migration. "
                "force option not used, and console input not available to confirm migration, aborting. "
                "Use the force option or run from an interactive shell to proceed with migration.",
                count
            )

    return confirmed


def _check_file_handle_exists(cursor, from_file_handle_id):
    # check if there is already a copied file handle for the given.
    # if so we can re-use it rather than making another copy
    row = cursor.execute(
        """
            select
                to_file_handle_id
            from
                migrations
            where
                from_file_handle_id = ?
                and to_file_handle_id is not null
        """,
        (from_file_handle_id,)
    ).fetchone()

    return row[0] if row else None


def migrate_indexed_files(
    syn: synapseclient.Synapse,
    db_path: str,
    create_table_snapshots=True,
    continue_on_error=False,
    force=False
):
    """
    Migrate files previously indexed in a sqlite database at the given db_path using the separate
    index_files_for_migration function. The files listed in the index will be migrated according to the
    configuration of that index.

    :param syn:                     A Synapse object with user's login, e.g. syn = synapseclient.login()
    :param db_path:                 A path on disk where a sqlite db was created using the index_files_for_migration
                                        function.
    :param create_table_snapshots:  When updating the files in any table, whether the a snapshot of the table is
                                        first created (default True).
    :param continue_on_error:       Whether any errors encountered while migrating will be raised
                                        or instead just recorded in the sqlite database while allowing the migration
                                        to continue. Default is False (any errors are raised).
    :param force:                   If running in an interactive shell, migration requires an interactice confirmation.
                                        This can be bypassed by using the force=True option.

    :return:                        A MigrationResult object that can be used to inspect the results of the migration.
    """
    executor, max_concurrent_file_copies = _get_executor(syn)

    test_import_sqlite3()
    import sqlite3
    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()

        _ensure_schema(cursor)
        settings = _retrieve_index_settings(cursor)
        if settings is None:
            # no settings were available at the index given
            raise ValueError(
                "Unable to retrieve existing index settings from '{}'. "
                "Either this path does represent a previously created migration index file or the file is corrupt."
            )

        dest_storage_location_id = settings['dest_storage_location_id']
        if not _confirm_migration(cursor, force, dest_storage_location_id):
            logging.info("Migration aborted.")
            return

        key = _MigrationKey(id='', type=None, row_id=-1, col_id=-1, version=-1)

        futures = set()

        # we keep track of the file handles that are currently being migrated
        # so that if we encounter multiple entities associated with the same
        # file handle we can copy the file handle once and update all the entities
        # with the single copied file handle
        pending_file_handle_ids = set()
        completed_file_handle_ids = set()

        # we keep track of the entity keys (syn id + version) so that we know
        # if we encounter the same one twice. normally we wouldn't but when we backtrack
        # to update any entities skipped because of a shared file handle we might
        # query for the same key as is already being operated on.
        pending_keys = set()

        batch_size = _get_batch_size()
        while True:
            # we query for additional file or table associated file handles to migrate in batches
            # ordering by synapse id. there can be multiple file handles associated with a particular
            # synapse id (i.e. multiple file entity versions or multiple table attached files per table),
            # so the ordering and where clause need to account for that.
            # we also include in the query any unmigrated files that were skipped previously through
            # the query loop that share a file handle with a file handle id that is now finished.
            version = key.version if key.version is not None else -1
            row_id = key.row_id if key.row_id is not None else -1
            col_id = key.col_id if key.col_id is not None else -1

            query_kwargs = {
                'indexed_status': _MigrationStatus.INDEXED.value,
                'id': key.id,
                'file_type': _MigrationType.FILE.value,
                'table_type': _MigrationType.TABLE_ATTACHED_FILE.value,
                'version': version,
                'row_id': row_id,
                'col_id': col_id,

                # ensure that we aren't ever adding more items to the shared executor than allowed
                'limit': min(batch_size, max_concurrent_file_copies - len(futures)),
            }

            # we can't use both named and positional literals in a query, so we use named
            # literals and then inline a string for the values for our file handle ids
            # since these are a dynamic list of values
            pending_file_handle_in = "('" + "','".join(pending_file_handle_ids) + "')"
            completed_file_handle_in = "('" + "','".join(completed_file_handle_ids) + "')"

            results = cursor.execute(
                f"""
                    select
                        id,
                        type,
                        version,
                        row_id,
                        col_id,
                        from_file_handle_id,
                        file_size
                    from migrations
                    where
                        status = :indexed_status
                        and (
                                (
                                    ((id > :id and type in (:file_type, :table_type))
                                    or (id = :id and type = :file_type and version is not null and version > :version)
                                    or (id = :id and type = :table_type and (row_id > :row_id or (row_id = :row_id and col_id > :col_id))))
                                    and from_file_handle_id not in {pending_file_handle_in}
                                ) or
                                (
                                    id <= :id 
                                    and from_file_handle_id in {completed_file_handle_in}
                                )
                        )
                    order by
                        id,
                        type,
                        row_id,
                        col_id,
                        version
                    limit :limit
                """,  # noqa
                query_kwargs,
            )

            row_count = 0
            for row in results:
                row_count += 1

                row_dict = _get_row_dict(cursor, row, True)
                key_dict = {
                    k: v for k, v in row_dict.items()
                    if k in ('id', 'type', 'version', 'row_id', 'col_id')
                }

                last_key = key
                key = _MigrationKey(**key_dict)
                from_file_handle_id = row_dict['from_file_handle_id']

                if key in pending_keys or from_file_handle_id in pending_file_handle_ids:
                    # if this record is already being migrated or it shares a file handle
                    # with a record that is being migrated then skip this.
                    # if it the record shares a file handle it will be picked up later
                    # when its file handle is completed.
                    continue

                file_size = row_dict['file_size']

                pending_keys.add(key)
                to_file_handle_id = _check_file_handle_exists(conn.cursor(), from_file_handle_id)
                if not to_file_handle_id:
                    pending_file_handle_ids.add(from_file_handle_id)

                if key.type == _MigrationType.FILE.value:
                    if key.version is None:
                        migration_fn = _create_new_file_version

                    else:
                        migration_fn = _migrate_file_version

                elif key.type == _MigrationType.TABLE_ATTACHED_FILE.value:
                    if last_key.id != key.id and create_table_snapshots:
                        syn.create_snapshot_version(key.id)

                    migration_fn = _migrate_table_attached_file

                else:
                    raise ValueError("Unexpected type {} with id {}".format(key.type, key.id))

                def migration_task(syn, key, from_file_handle_id, to_file_handle_id, file_size, storage_location_id):
                    # a closure to wrap the actual function call so that we an add some local variables
                    # to the return tuple which will be consumed when the future is processed
                    with shared_executor(executor):
                        try:
                            # instrument the shared executor in this thread so that we won't
                            # create a new executor to perform the multipart copy
                            to_file_handle_id = migration_fn(
                                syn,
                                key,
                                from_file_handle_id,
                                to_file_handle_id,
                                file_size,
                                storage_location_id)
                            return key, from_file_handle_id, to_file_handle_id
                        except Exception as ex:
                            raise _MigrationError(key, from_file_handle_id, to_file_handle_id) from ex

                future = executor.submit(
                    migration_task,
                    syn,
                    key,
                    from_file_handle_id,
                    to_file_handle_id,
                    file_size,
                    dest_storage_location_id,
                )
                futures.add(future)

            if row_count == 0 and not pending_file_handle_ids:
                # we've run out of migratable sqlite rows, we have nothing else
                # to submit, so we break out and wait for all remaining
                # tasks to conclude.
                break

            if len(futures) >= max_concurrent_file_copies or row_count < batch_size:
                # if we have no concurrency left to process any additional entities
                # or if we're near the end of he migration and have a small
                # remainder batch then we wait for one of the processing migrations
                # to finish. a small batch doesn't mean this is the last batch since
                # a completed file handle here could be associated with another
                # entity that we deferred before because it shared the same file handle id
                futures, completed_file_handle_ids = _wait_futures(
                    conn,
                    cursor,
                    futures,
                    pending_keys,
                    concurrent.futures.FIRST_COMPLETED,
                    continue_on_error,
                )

                pending_file_handle_ids -= completed_file_handle_ids

        if futures:
            # wait for all remaining migrations to conclude before returning
            _wait_futures(
                conn,
                cursor,
                futures,
                pending_keys,
                concurrent.futures.ALL_COMPLETED,
                continue_on_error
            )

    return MigrationResult(syn, db_path)


def _verify_storage_location_ownership(syn, storage_location_id):
    # if this doesn't raise an error we're okay
    try:
        syn.restGET("/storageLocation/{}".format(storage_location_id))
    except synapseclient.core.exceptions.SynapseHTTPError:
        raise ValueError(
            "Error verifying storage location ownership of {}. You must be creator of the destination storage location"
            .format(storage_location_id)
        )


def _retrieve_index_settings(cursor):
    # index settings are stored as a json-string in a one-row table
    import sqlite3
    settings = None
    try:
        results = cursor.execute("select settings from migration_settings")
        row = results.fetchone()
        if row:
            settings = json.loads(row[0])
    except (sqlite3.OperationalError, ValueError) as ex:
        raise ValueError(
            "Unable to parse index settings, the index may be corrupt or created by an older version "
            "of this function. You will need to re-create the index."
        ) from ex

    return settings


def _verify_index_settings(
        cursor,
        db_path,
        root_id,
        dest_storage_location_id,
        source_storage_location_ids,
        file_version_strategy,
        include_table_files,
):
    existing_settings = _retrieve_index_settings(cursor)

    if existing_settings is not None:
        settings = locals()
        for setting in (
            'root_id',
            'dest_storage_location_id',
            'source_storage_location_ids',
            'file_version_strategy',
            'include_table_files',
        ):
            parameter = settings[setting]
            existing_value = existing_settings[setting]

            if not parameter == existing_value:
                # value does not match the existing index settings.
                # we can't resume indexing with an existing index file using a different setting.
                raise ValueError(
                    "Index parameter does not match the setting recorded in the existing index file. "
                    "To change the index settings start over by deleting the file or using a different path. "
                    "Expected {} '{}', found '{}' in index file '{}'".format(
                        setting,
                        existing_value,
                        parameter,
                        db_path,
                    )
                )

    else:
        # this is a new index file, no previous values to compare against,
        # instead record the current settings
        settings_str = json.dumps({
            'root_id': root_id,
            'dest_storage_location_id': dest_storage_location_id,
            'source_storage_location_ids': source_storage_location_ids,
            'file_version_strategy': file_version_strategy,
            'include_table_files': 1 if include_table_files else 0,
        })
        cursor.execute("insert into migration_settings (settings) values (?)", (settings_str,))


def _check_indexed(cursor, entity_id):
    # check if we have indexed the given entity in the sqlite db yet.
    # if so it can skip reindexing it. supports resumption.
    indexed_row = cursor.execute(
        "select 1 from migrations where id = ?",
        (entity_id,)
    ).fetchone()

    if indexed_row:
        logging.debug('%s already indexed, skipping', entity_id)
        return True

    logging.debug('%s not yet indexed, indexing now', entity_id)
    return False


def _get_version_numbers(syn, entity_id):
    for version_info in syn._GET_paginated("/entity/{id}/version".format(id=entity_id)):
        yield version_info['versionNumber']


def _include_file_storage_location_in_index(
    file_handle,
    source_storage_location_ids,
    to_storage_location_id,
):
    # helper determines whether a file is included in the index depending on its storage location.
    # if source_storage_location_ids are specified the from storage location must be in it.
    # if the current storage location already matches the destination location then we also
    # include it in the index, we'll mark it as already migrated.

    from_storage_location_id = file_handle.get('storageLocationId')
    if (
            (file_handle.get('concreteType') == concrete_types.S3_FILE_HANDLE) and
            (
                not source_storage_location_ids or
                str(from_storage_location_id) in source_storage_location_ids or
                str(from_storage_location_id) == str(to_storage_location_id)
            )
    ):
        migration_status = _MigrationStatus.INDEXED.value \
            if str(from_storage_location_id) != str(to_storage_location_id) \
            else _MigrationStatus.ALREADY_MIGRATED.value
        return migration_status

    # this file is not included in this index
    return None


def _index_file_entity(
        cursor,
        syn,
        entity_id,
        parent_id,
        to_storage_location_id,
        source_storage_location_ids,
        file_version_strategy
):
    logging.info('Indexing file entity %s', entity_id)

    # 2-tuples of entity, version # to record
    entity_versions = []

    if file_version_strategy == 'new':
        # we'll need the etag to be able to do an update on an entity version
        # so we need to fetch the full entity now
        entity = syn.get(entity_id, downloadFile=False)
        entity_versions.append((
            entity,
            None  # no version number, this record indicates we will create a new version
        ))

    elif file_version_strategy == 'all':
        # one row for each existing version that will all be migrated
        for version in _get_version_numbers(syn, entity_id):
            entity = syn.get(entity_id, version=version, downloadFile=False)
            entity_versions.append((
                entity,
                version)
            )

    elif file_version_strategy == 'latest':
        # one row for the most recent version that will be migrated
        entity = syn.get(entity_id, downloadFile=False)
        entity_versions.append((
            entity,
            entity.versionNumber
        ))

    if entity_versions:
        insert_values = []
        for (entity, version) in entity_versions:
            migration_status = _include_file_storage_location_in_index(
                entity._file_handle,
                source_storage_location_ids,
                to_storage_location_id
            )
            if migration_status:

                file_size = entity._file_handle['contentSize']
                insert_values.append((
                    entity_id,
                    _MigrationType.FILE.value,
                    version,
                    parent_id,
                    entity._file_handle['storageLocationId'],
                    entity.dataFileHandleId,
                    file_size,
                    migration_status
                ))

        if insert_values:
            cursor.executemany(
                """
                    insert into migrations (
                        id,
                        type,
                        version,
                        parent_id,
                        from_storage_location_id,
                        from_file_handle_id,
                        file_size,
                        status
                    ) values (?, ?, ?, ?, ?, ?, ?, ?)
                """,
                insert_values
            )


def _get_table_file_handle_rows(syn, table_id):
    file_handle_columns = [c for c in syn.getTableColumns(table_id) if c['columnType'] == 'FILEHANDLEID']
    if file_handle_columns:
        file_column_select = join_column_names(file_handle_columns)
        results = syn.tableQuery("select {} from {}".format(file_column_select, table_id))
        for row in results:
            file_handles = {}

            # first two cols are row id and row version, rest are file handle ids from our query
            row_id, row_version = row[:2]

            file_handle_ids = row[2:]
            for i, file_handle_id in enumerate(file_handle_ids):
                if file_handle_id:
                    col_id = file_handle_columns[i]['id']
                    file_handle = syn._getFileHandleDownload(
                        file_handle_id,
                        table_id,
                        objectType='TableEntity'
                    )['fileHandle']
                    file_handles[col_id] = file_handle

            yield row_id, row_version, file_handles


def _index_table_entity(
        cursor,
        syn,
        entity,
        parent_id,
        dest_storage_location_id,
        source_storage_location_ids
):
    entity_id = utils.id_of(entity)
    logging.info('Indexing table entity %s', entity_id)

    row_batch = []

    def _insert_row_batch(row_batch):
        if row_batch:
            cursor.executemany(
                """insert into migrations
                    (
                        id,
                        type,
                        parent_id,
                        row_id,
                        col_id,
                        version,
                        from_storage_location_id,
                        from_file_handle_id,
                        file_size,
                        status
                    ) values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                row_batch
            )

    for row_id, row_version, file_handles in _get_table_file_handle_rows(syn, entity_id):
        for col_id, file_handle in file_handles.items():
            migration_status = _include_file_storage_location_in_index(
                file_handle,
                source_storage_location_ids,
                dest_storage_location_id,
            )
            if migration_status:
                file_size = file_handle['contentSize']
                row_batch.append((
                    entity_id,
                    _MigrationType.TABLE_ATTACHED_FILE.value,
                    parent_id,
                    row_id,
                    col_id,
                    row_version,
                    file_handle['storageLocationId'],
                    file_handle['id'],
                    file_size,
                    migration_status
                ))

                if len(row_batch) % _get_batch_size() == 0:
                    _insert_row_batch(row_batch)
                    row_batch = []

    if row_batch:
        _insert_row_batch(row_batch)


def _index_container(
        conn,
        cursor,
        syn,
        container_entity,
        parent_id,
        dest_storage_location_id,
        source_storage_location_ids,
        file_version_strategy,
        include_table_files,
        continue_on_error
):
    entity_id = utils.id_of(container_entity)
    concrete_type = utils.concrete_type_of(container_entity)
    logging.info('Indexing %s %s', concrete_type[concrete_type.rindex('.') + 1:], entity_id)

    include_types = []
    if file_version_strategy != 'skip':
        include_types.extend(('folder', 'file'))
    if include_table_files:
        include_types.append('table')

    children = syn.getChildren(entity_id, includeTypes=include_types)
    for child in children:
        _index_entity(
            conn,
            cursor,
            syn,
            child,
            entity_id,
            dest_storage_location_id,
            source_storage_location_ids,
            file_version_strategy,
            include_table_files,
            continue_on_error,
        )

    # once all the children are recursively indexed we mark this parent itself as indexed
    container_type = (
        _MigrationType.PROJECT.value
        if concrete_types.PROJECT_ENTITY == concrete_type
        else _MigrationType.FOLDER.value
    )
    cursor.execute(
        "insert into migrations (id, type, parent_id, status) values (?, ?, ?, ?)",
        [entity_id, container_type, parent_id, _MigrationStatus.INDEXED.value]
    )


def _index_entity(
        conn,
        cursor,
        syn,
        entity,
        parent_id,
        dest_storage_location_id,
        source_storage_location_ids,
        file_version_strategy,
        include_table_files,
        continue_on_error
):
    # recursive function to index a given entity into the sqlite db.

    entity_id = utils.id_of(entity)
    concrete_type = utils.concrete_type_of(entity)

    try:
        if not _check_indexed(cursor, entity_id):
            # if already indexed we short circuit (previous indexing will be used)

            if concrete_type == concrete_types.FILE_ENTITY:
                _index_file_entity(
                    cursor,
                    syn,
                    entity_id,
                    parent_id,
                    dest_storage_location_id,
                    source_storage_location_ids,
                    file_version_strategy,
                )

            elif concrete_type == concrete_types.TABLE_ENTITY:
                _index_table_entity(
                    cursor,
                    syn,
                    entity,
                    parent_id,
                    dest_storage_location_id,
                    source_storage_location_ids,
                )

            elif concrete_type in [concrete_types.FOLDER_ENTITY, concrete_types.PROJECT_ENTITY]:
                _index_container(
                    conn,
                    cursor,
                    syn,
                    entity,
                    parent_id,
                    dest_storage_location_id,
                    source_storage_location_ids,
                    file_version_strategy,
                    include_table_files,
                    continue_on_error,
                )

        conn.commit()

    except _IndexingError:
        # this is a recursive function, we don't need to log the error at every level so just
        # pass up exceptions of this type that wrap the underlying exception and indicate
        # that they were already logged
        raise

    except Exception as ex:

        if continue_on_error:
            logging.warning("Error indexing entity %s of type %s", entity_id, concrete_type, exc_info=True)
            tb_str = ''.join(traceback.format_exception(type(ex), ex, ex.__traceback__))

            cursor.execute(
                """
                    insert into migrations (
                        id,
                        type,
                        parent_id,
                        status,
                        exception
                    ) values (?, ?, ?, ?, ?)
                """,
                (
                    entity_id,
                    _MigrationType.from_concrete_type(concrete_type).value,
                    parent_id,
                    _MigrationStatus.ERRORED.value,
                    tb_str,
                )
            )

        else:
            raise _IndexingError(entity_id, concrete_type) from ex


def _get_part_size(file_size):
    return max(DEFAULT_PART_SIZE, math.ceil((file_size / MAX_NUMBER_OF_PARTS)))


def _create_new_file_version(syn, key, from_file_handle_id, to_file_handle_id, file_size, storage_location_id):
    logging.info('Creating new version for file entity %s', key.id)

    entity = syn.get(key.id, downloadFile=False)

    source_file_handle_association = {
        'fileHandleId': from_file_handle_id,
        'associateObjectId': key.id,
        'associateObjectType': 'FileEntity',
    }

    # copy to a new file handle if we haven't already
    if not to_file_handle_id:
        to_file_handle_id = multipart_copy(
            syn,
            source_file_handle_association,
            storage_location_id=storage_location_id,
            part_size=_get_part_size(file_size)
        )

    entity.dataFileHandleId = to_file_handle_id
    syn.store(entity)

    return to_file_handle_id


def _migrate_file_version(syn, key, from_file_handle_id, to_file_handle_id, file_size, storage_location_id):
    logging.info('Migrating file entity %s version %s', key.id, key.version)

    source_file_handle_association = {
        'fileHandleId': from_file_handle_id,
        'associateObjectId': key.id,
        'associateObjectType': 'FileEntity',
    }

    # copy to a new file handle if we haven't already
    if not to_file_handle_id:
        to_file_handle_id = multipart_copy(
            syn,
            source_file_handle_association,
            storage_location_id=storage_location_id,
            part_size=_get_part_size(file_size),
        )

    file_handle_update_request = {
        'oldFileHandleId': from_file_handle_id,
        'newFileHandleId': to_file_handle_id,
    }

    # no response, we rely on a 200 here
    syn.restPUT(
        "/entity/{id}/version/{versionNumber}/filehandle".format(
            id=key.id,
            versionNumber=key.version,
        ),
        json.dumps(file_handle_update_request),
    )

    return to_file_handle_id


def _migrate_table_attached_file(syn, key, from_file_handle_id, to_file_handle_id, file_size, storage_location_id):
    logging.info('Migrating table attached file %s, row %s, col %s', key.id, key.row_id, key.col_id)

    source_file_handle_association = {
        'fileHandleId': from_file_handle_id,
        'associateObjectId': key.id,
        'associateObjectType': 'TableEntity',
    }

    # copy to a new file handle if we haven't already
    if not to_file_handle_id:
        to_file_handle_id = multipart_copy(
            syn,
            source_file_handle_association,
            storage_location_id=storage_location_id,
            part_size=_get_part_size(file_size),
        )

    row_mapping = {str(key.col_id): to_file_handle_id}
    partial_rows = [synapseclient.table.PartialRow(row_mapping, key.row_id)]
    partial_rowset = synapseclient.PartialRowset(key.id, partial_rows)
    syn.store(partial_rowset)

    return to_file_handle_id


class _MigrationError(Exception):
    def __init__(self, key, from_file_handle_id, to_file_handle_id):
        self.key = key
        self.from_file_handle_id = from_file_handle_id
        self.to_file_handle_id = to_file_handle_id


class _IndexingError(Exception):
    def __init__(self, entity_id, concrete_type):
        self.entity_id = entity_id
        self.concrete_type = concrete_type



"""
***********
Annotations
***********

Annotations are arbitrary metadata attached to Synapse entities. They can be accessed like ordinary object properties
or like dictionary keys::

    entity.my_annotation = 'This is one way to do it'
    entity['other_annotation'] = 'This is another'

Annotations can be given in the constructor for Synapse Entities::

    entity = File('data.xyz', parent=my_project, rating=9.1234)

Annotate the entity with location data::

    entity.lat_long = [47.627477, -122.332154]

Record when we collected the data::

    from datetime import datetime as Datetime
    entity.collection_date = Datetime.now()

See:

- :py:meth:`synapseclient.Synapse.get_annotations`
- :py:meth:`synapseclient.Synapse.set_annotations`

~~~~~~~~~~~~~~~~~~~~~~~
Annotating data sources
~~~~~~~~~~~~~~~~~~~~~~~

Data sources are best recorded using Synapse's `provenance <Activity.html>`_ tools.

~~~~~~~~~~~~~~~~~~~~~~
Implementation details
~~~~~~~~~~~~~~~~~~~~~~

In Synapse, entities have both properties and annotations. Properties are used by the system, whereas annotations are
completely user defined. In the Python client, we try to present this situation as a normal object, with one set of
properties.

For more on the implementation and a few gotchas, see the documentation on :py:mod:`synapseclient.entity`.

See also:

- :py:class:`synapseclient.entity.Entity`
- :py:mod:`synapseclient.entity`


~~~~~~~
Classes
~~~~~~~
.. autoclass:: synapseclient.annotations.Annotations
    :members:

    .. automethod:: __init__



"""

import collections

from .entity import Entity
from synapseclient.core.utils import to_unix_epoch_time, from_unix_epoch_time, is_date, to_list, id_of
import typing
import datetime


def _identity(x):
    return x


def raise_anno_type_error(anno_type: str):
    raise ValueError(f"Unknown type in annotations response: {anno_type}")


ANNO_TYPE_TO_FUNC: typing.Dict[str, typing.Callable[[str], typing.Union[str, int, float, datetime.datetime]]] = \
    collections.defaultdict(
        raise_anno_type_error,
        {
            'STRING': _identity,
            'BOOLEAN': lambda bool_str: bool_str == 'true',
            'LONG': int,
            'DOUBLE': float,
            'TIMESTAMP_MS': lambda time_str: from_unix_epoch_time(int(time_str))
        }
    )


def is_synapse_annotations(annotations: typing.Mapping):
    """Tests if the given object is a Synapse-style Annotations object."""
    if not isinstance(annotations, collections.abc.Mapping):
        return False
    return annotations.keys() >= {'id', 'etag', 'annotations'}


def _annotation_value_list_element_type(annotation_values: typing.List):
    if not annotation_values:
        raise ValueError("annotations value list can not be empty")

    first_element_type = type(annotation_values[0])

    if all(isinstance(x, first_element_type) for x in annotation_values):
        return first_element_type

    return object


def is_submission_status_annotations(annotations):
    """Tests if the given dictionary is in the form of annotations to submission status"""
    keys = ['objectId', 'scopeId', 'stringAnnos', 'longAnnos', 'doubleAnnos']
    if not isinstance(annotations, collections.abc.Mapping):
        return False
    return all([key in keys for key in annotations.keys()])


def to_submission_status_annotations(annotations, is_private=True):
    """
    Converts a normal dictionary to the format used to annotate submission statuses, which is different from the format
    used to annotate entities.

    :param annotations: A normal Python dictionary whose values are strings, floats, ints or doubles

    :param is_private: Set privacy on all annotations at once. These can be set individually using
                       :py:func:`set_privacy`.

    Example::

        from synapseclient.annotations import to_submission_status_annotations, from_submission_status_annotations
        from datetime import datetime as Datetime

        ## create a submission and get its status
        submission = syn.submit(evaluation, 'syn11111111')
        submission_status = syn.getSubmissionStatus(submission)

        ## add annotations
        submission_status.annotations = {'foo':'bar', 'shoe_size':12, 'IQ':12, 'timestamp':Datetime.now()}

        ## convert annotations
        submission_status.annotations = to_submission_status_annotations(submission_status.annotations)
        submission_status = syn.store(submission_status)


    Synapse categorizes these annotations by: stringAnnos, doubleAnnos, longAnnos.
    """
    if is_submission_status_annotations(annotations):
        return annotations
    synapseAnnos = {}
    for key, value in annotations.items():
        if key in ['objectId', 'scopeId', 'stringAnnos', 'longAnnos', 'doubleAnnos']:
            synapseAnnos[key] = value
        elif isinstance(value, bool):
            synapseAnnos.setdefault('stringAnnos', []) \
                .append({'key': key, 'value': str(value).lower(), 'isPrivate': is_private})
        elif isinstance(value, int):
            synapseAnnos.setdefault('longAnnos', []) \
                .append({'key': key, 'value': value, 'isPrivate': is_private})
        elif isinstance(value, float):
            synapseAnnos.setdefault('doubleAnnos', []) \
                .append({'key': key, 'value': value, 'isPrivate': is_private})
        elif isinstance(value, str):
            synapseAnnos.setdefault('stringAnnos', []) \
                .append({'key': key, 'value': value, 'isPrivate': is_private})
        elif is_date(value):
            synapseAnnos.setdefault('longAnnos', []) \
                .append({'key': key, 'value': to_unix_epoch_time(value), 'isPrivate': is_private})
        else:
            synapseAnnos.setdefault('stringAnnos', []) \
                .append({'key': key, 'value': str(value), 'isPrivate': is_private})
    return synapseAnnos


# TODO: this should accept a status object and return its annotations or an empty dict if there are none
def from_submission_status_annotations(annotations):
    """
    Convert back from submission status annotation format to a normal dictionary.

    Example::

        submission_status.annotations = from_submission_status_annotations(submission_status.annotations)
    """
    dictionary = {}
    for key, value in annotations.items():
        if key in ['stringAnnos', 'longAnnos']:
            dictionary.update({kvp['key']: kvp['value'] for kvp in value})
        elif key == 'doubleAnnos':
            dictionary.update({kvp['key']: float(kvp['value']) for kvp in value})
        else:
            dictionary[key] = value
    return dictionary


def set_privacy(annotations, key, is_private=True, value_types=['longAnnos', 'doubleAnnos', 'stringAnnos']):
    """
    Set privacy of individual annotations, where annotations are in the format used by Synapse SubmissionStatus objects.
    See the `Annotations documentation \
    <http://docs.synapse.org/rest/org/sagebionetworks/repo/model/annotation/Annotations.html>`_ and the docs regarding
    `querying annotations <http://docs.synapse.org/rest/GET/evaluation/submission/query.html>`_.

    :param annotations: Annotations that have already been converted to Synapse format using
                        :py:func:`to_submission_status_annotations`.
    :param key:         The key of the annotation whose privacy we're setting.
    :param is_private:  If False, the annotation will be visible to users with READ permission on the evaluation.
                        If True, the it will be visible only to users with READ_PRIVATE_SUBMISSION on the evaluation.
                        Note: Is this really correct???
    :param value_types: A list of the value types in which to search for the key. Defaults to all types
                        ['longAnnos', 'doubleAnnos', 'stringAnnos'].

    """
    for value_type in value_types:
        kvps = annotations.get(value_type, None)
        if kvps:
            for kvp in kvps:
                if kvp['key'] == key:
                    kvp['isPrivate'] = is_private
                    return kvp
    raise KeyError('The key "%s" couldn\'t be found in the annotations.' % key)


class Annotations(dict):
    """
     Represent Synapse Entity annotations as a flat dictionary with the system assigned properties id, etag
     as object attributes.
    """
    id: str
    etag: str

    def __init__(self, id: typing.Union[str, int, Entity], etag: str, values: typing.Dict = None, **kwargs):
        """
        Create an Annotations object taking key value pairs from a dictionary or from keyword arguments.
        System properties id, etag, creationDate and uri become attributes of the object.

        :param id:  A Synapse ID, a Synapse Entity object, a plain dictionary in which 'id' maps to a Synapse ID
        :param etag: etag of the Synapse Entity
        :param values:  (Optional) dictionary of values to be copied into annotations

        :param **kwargs: additional key-value pairs to be added as annotations

        Example::

            example1 = Annotations('syn123','40256475-6fb3-11ea-bb0a-9cb6d0d8d984', {'foo':'bar'})

            example2 = Annotations('syn123','40256475-6fb3-11ea-bb0a-9cb6d0d8d984', foo='bar')


            example3 = Annotations('syn123','40256475-6fb3-11ea-bb0a-9cb6d0d8d984')
            example3['foo'] = 'bar'

        """
        super().__init__()

        self.id = id
        self.etag = etag

        if values:
            self.update(values)
        if kwargs:
            self.update(kwargs)

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if value is None:
            raise ValueError("id must not be None")
        self._id = id_of(value)

    @property
    def etag(self):
        return self._etag

    @etag.setter
    def etag(self, value):
        if value is None:
            raise ValueError("etag must not be None")
        self._etag = str(value)


def to_synapse_annotations(annotations: Annotations) -> typing.Dict[str, typing.Any]:
    """Transforms a simple flat dictionary to a Synapse-style Annotation object.
    https://rest-docs.synapse.org/rest/org/sagebionetworks/repo/model/annotation/v2/Annotations.html
    """

    if is_synapse_annotations(annotations):
        return annotations
    synapse_annos = {}

    if not isinstance(annotations, Annotations):
        raise TypeError("annotations must be a synapseclient.Annotations object with 'id' and 'etag' attributes")

    synapse_annos['id'] = annotations.id
    synapse_annos['etag'] = annotations.etag

    synapse_annos['annotations'] = _convert_to_annotations_list(annotations)
    return synapse_annos


def _convert_to_annotations_list(annotations):
    nested_annos = {}
    for key, value in annotations.items():
        elements = to_list(value)
        element_cls = _annotation_value_list_element_type(elements)
        if issubclass(element_cls, str):
            nested_annos[key] = {'type': 'STRING',
                                 'value': elements}
        elif issubclass(element_cls, bool):
            nested_annos[key] = {'type': 'BOOLEAN',
                                 'value': ['true' if e else 'false' for e in elements]}
        elif issubclass(element_cls, int):
            nested_annos[key] = {'type': 'LONG',
                                 'value': [str(e) for e in elements]}
        elif issubclass(element_cls, float):
            nested_annos[key] = {'type': 'DOUBLE',
                                 'value': [str(e) for e in elements]}
        elif issubclass(element_cls, (datetime.date, datetime.datetime)):
            nested_annos[key] = {'type': 'TIMESTAMP_MS',
                                 'value': [str(to_unix_epoch_time(e)) for e in elements]}
        else:
            nested_annos[key] = {'type': 'STRING',
                                 'value': [str(e) for e in elements]}
    return nested_annos


def from_synapse_annotations(raw_annotations: typing.Dict[str, typing.Any]) -> Annotations:
    """Transforms a Synapse-style Annotation object to a simple flat dictionary."""
    if not is_synapse_annotations(raw_annotations):
        raise ValueError(
            'Unexpected format of annotations from Synapse. Must include keys: "id", "etag", and "annotations"')

    annos = Annotations(raw_annotations['id'], raw_annotations['etag'])
    for key, value_and_type in raw_annotations['annotations'].items():
        key: str
        conversion_func = ANNO_TYPE_TO_FUNC[value_and_type['type']]
        annos[key] = [conversion_func(v) for v in value_and_type['value']]

    return annos


def check_annotations_changed(bundle_annotations, new_annotations):
    converted_annos = _convert_to_annotations_list(new_annotations)
    return bundle_annotations['annotations'] != converted_annos


def convert_old_annotation_json(annotations):
    """Transforms a parsed JSON dictionary of old style annotations
    into a new style consistent with the entity bundle v2 format.

    This is intended to support some models that were saved as serialized
    entity bundle JSON (Submissions). we don't need to support newer
    types here e.g. BOOLEAN because they did not exist at the time
    that annotation JSON was saved in this form.
    """

    meta_keys = ('id', 'etag', 'creationDate', 'uri')

    type_mapping = {
        'doubleAnnotations': 'DOUBLE',
        'stringAnnotations': 'STRING',
        'longAnnotations': "LONG",
        'dateAnnotations': 'TIMESTAMP_MS',
    }

    annos_v1_keys = set(meta_keys) | set(type_mapping.keys())

    # blobAnnotations appear to be little/unused and there is no mapping defined here but if they
    # are present on the annos we should treat it as an old style annos dict
    annos_v1_keys.add('blobAnnotations')

    # if any keys in the annos dict are not consistent with an old style annotations then we treat
    # it as an annotations2 style dictionary that is not in need of any conversion
    if any(k not in annos_v1_keys for k in annotations.keys()):
        return annotations

    converted = {k: v for k, v in annotations.items() if k in meta_keys}
    converted_annos = converted['annotations'] = {}

    for old_type_key, converted_type in type_mapping.items():
        values = annotations.get(old_type_key)
        if values:
            for k, vs in values.items():
                converted_annos[k] = {
                    'type': converted_type,
                    'value': vs,
                }

    return converted



import abc
import os

import deprecated.sphinx

from synapseclient.core.credentials import cached_sessions
from synapseclient.core.credentials.cred_data import SynapseApiKeyCredentials, SynapseAuthTokenCredentials
from synapseclient.core.exceptions import SynapseAuthenticationError


class SynapseCredentialsProvider(metaclass=abc.ABCMeta):
    """
    A credential provider is responsible for retrieving synapse authentication information (e.g. username/password or
    username/api key) from a source(e.g. login args, config file, cached credentials in keyring), and use them to return
    a ``SynapseCredentials` instance.
    """

    @abc.abstractmethod
    def _get_auth_info(self, syn, user_login_args):
        """
        Subclasses must implement this to decide how to obtain username, password, and api_key.
        For any of these 3 values, return None if it is not possible to get that value.

        Not all implementations will need to make use of the user_login_args parameter or syn.
        These parameters provide context about the Synapse client's configuration and login() arguments.

        :param ``synapseclient.client.Synapse`` syn:        Synapse client instance
        :param ``cred_data.UserLoginArgs`` user_login_args: subset of arguments passed during syn.login()
        :return: tuple of (username, password, api_key, bearer auth token e.g. a personal access token),
                    any of these values could None if it is not available.
        """
        return None, None, None, None

    def get_synapse_credentials(self, syn, user_login_args):
        """
        Returns `SynapseCredentials` if this provider is able to get valid credentials, returns None otherwise.
        :param ``synapseclient.client.Synapse`` syn:        Synapse client instance
        :param ``cred_data.UserLoginArgs`` user_login_args: subset of arguments passed during syn.login()
        :return: `SynapseCredentials` if valid credentials can be found by this provider, None otherwise
        """
        return self._create_synapse_credential(syn, *self._get_auth_info(syn, user_login_args))

    def _create_synapse_credential(self, syn, username, password, api_key, auth_token):
        if username is not None:
            if password is not None:
                retrieved_session_token = syn._getSessionToken(email=username, password=password)
                return SynapseApiKeyCredentials(syn._getAPIKey(retrieved_session_token), username)

            elif auth_token is None and api_key is not None:
                # auth token takes precedence over api key
                return SynapseApiKeyCredentials(api_key, username)

        if auth_token is not None:
            credentials = SynapseAuthTokenCredentials(auth_token)
            profile = syn.restGET('/userProfile', auth=credentials)
            profile_username = profile.get('userName')
            profile_emails = profile.get('emails', [])

            if username and (username != profile_username and username not in profile_emails):
                # a username/email is not required when logging in with an auth token however if both are provided
                # raise an error if they do not correspond to avoid any ambiguity about what profile was logged in
                raise SynapseAuthenticationError(
                    'username/email and auth_token both provided but username does not match token profile'
                )

            credentials.username = profile_username
            return credentials

        return None


class UserArgsCredentialsProvider(SynapseCredentialsProvider):
    """
    Retrieves auth info from user_login_args
    """

    def _get_auth_info(self, syn, user_login_args):
        return (
            user_login_args.username,
            user_login_args.password,
            user_login_args.api_key,
            user_login_args.auth_token,
        )


@deprecated.sphinx.deprecated(version='1.9.0', action='ignore',
                              reason="This will be removed in 2.0. Please use username and password or apiKey instead.")
class UserArgsSessionTokenCredentialsProvider(SynapseCredentialsProvider):
    """
    This is a special case where we are not given context as to what the username is. We are only given a session token
    and must retrieve the username and api key from Synapse
    """

    def _get_auth_info(self, syn, user_login_args):
        username = None
        password = None
        api_key = None
        auth_token = None

        if user_login_args.session_token:
            username = syn.getUserProfile(sessionToken=user_login_args.session_token)['userName']
            api_key = syn._getAPIKey(user_login_args.session_token)

        return username, password, api_key, auth_token


class ConfigFileCredentialsProvider(SynapseCredentialsProvider):
    """
    Retrieves auth info from .synapseConfig file
    """

    def _get_auth_info(self, syn, user_login_args):
        config_dict = syn._get_config_authentication()
        # check to make sure we didn't accidentally provide the wrong user

        username = config_dict.get('username')
        password = config_dict.get('password')
        api_key = config_dict.get('apikey')
        token = config_dict.get('authtoken')

        if user_login_args.username and username != user_login_args.username:
            # if the username is provided and there is a config file username but they don't match
            # then we don't use any of the values from the config to prevent ambiguity
            username = None
            password = None
            api_key = None
            token = None

        return username, password, api_key, token


class CachedCredentialsProvider(SynapseCredentialsProvider):
    """
    Retrieves auth info from cached_sessions
    """

    def _get_auth_info(self, syn, user_login_args):
        username = None
        password = None
        api_key = None
        auth_token = None

        if not user_login_args.skip_cache:
            username = user_login_args.username or cached_sessions.get_most_recent_user()
            if username:
                api_creds = SynapseApiKeyCredentials.get_from_keyring(username)
                auth_token_creds = SynapseAuthTokenCredentials.get_from_keyring(username)

                api_key = api_creds.secret if api_creds else None
                auth_token = auth_token_creds.secret if auth_token_creds else None

        return username, password, api_key, auth_token


class AWSParameterStoreCredentialsProvider(SynapseCredentialsProvider):
    """
    Retrieves user's authentication token from AWS SSM Parameter store
    """

    ENVIRONMENT_VAR_NAME = "SYNAPSE_TOKEN_AWS_SSM_PARAMETER_NAME"

    def _get_auth_info(self, syn, user_login_args):
        ssm_param_name = os.environ.get(self.ENVIRONMENT_VAR_NAME)
        token = None
        if ssm_param_name:
            try:
                import boto3
                import botocore
                ssm_client = boto3.client('ssm')
                result = ssm_client.get_parameter(
                    Name=ssm_param_name,
                    WithDecryption=True,
                )
                token = result['Parameter']['Value']
            except ImportError:
                syn.logger.warning(
                    f'{self.ENVIRONMENT_VAR_NAME} was defined as {ssm_param_name}, but "boto3" could not be imported.'
                    ' The Synapse client uses "boto3" in order to access Systems Manager Parameter Storage.'
                    ' Please ensure that you have installed "boto3" to enable this feature.')
            # this except block must be defined after the ImportError except block
            # otherwise, there's no guarantee "botocore" is already imported and defined
            except botocore.exceptions.ClientError:
                syn.logger.warning(f'{self.ENVIRONMENT_VAR_NAME} was defined as {ssm_param_name}, '
                                   'but the matching parameter name could not be found in AWS Parameter Store. '
                                   f'Caused by AWS error:\n', exc_info=True)

        # if username is included in user's arguments, return it so that
        # it may be validated against the username authenticated by the token
        return user_login_args.username, None, None, token


class EnvironmentVariableCredentialsProvider(SynapseCredentialsProvider):
    """
    Retrieves the user's auth token from an environment variable
    """
    ENVIRONMENT_VAR_NAME = "SYNAPSE_AUTH_TOKEN"

    def _get_auth_info(self, syn, user_login_args):
        return user_login_args.username, None, None, os.environ.get(self.ENVIRONMENT_VAR_NAME)


class SynapseCredentialsProviderChain(object):
    """
    Class that has a list of ``SynapseCredentialsProvider`` from which this class attempts to retrieve
    ``SynapseCredentials``.
    """

    def __init__(self, cred_providers):
        """
        :param list[``SynapseCredentialsProvider``] cred_providers: list of credential providers
        """
        self.cred_providers = list(cred_providers)

    def get_credentials(self, syn, user_login_args):
        """
        Iterates its list of ``SynapseCredentialsProvider`` and returns the first non-None ``SynapseCredential``
        returned by a provider. If no provider is able to provide a ``SynapseCredential``, returns None.
        :param ``synapseclient.client.Synapse`` syn:        Synapse client instance
        :param ``cred_data.UserLoginArgs`` user_login_args: subset of arguments passed during syn.login()
        :return: `SynapseCredentials` returned by the first non-None provider in its list, None otherwise
        """
        for provider in self.cred_providers:
            creds = provider.get_synapse_credentials(syn, user_login_args)
            if creds is not None:
                return creds
        return None


# NOTE: If you change the order of this list, please also change the documentation in Synapse.login() that describes the
# order

DEFAULT_CREDENTIAL_PROVIDER_CHAIN = SynapseCredentialsProviderChain([
    UserArgsSessionTokenCredentialsProvider(),  # This provider is DEPRECATED
    UserArgsCredentialsProvider(),
    EnvironmentVariableCredentialsProvider(),
    ConfigFileCredentialsProvider(),
    CachedCredentialsProvider(),
    AWSParameterStoreCredentialsProvider(),  # see service catalog issue: SC-260
])


def get_default_credential_chain():
    """
    :return: credential chain
    :rtype: ```SynapseCredentialsProviderChain``
    """
    return DEFAULT_CREDENTIAL_PROVIDER_CHAIN



"""
**********
Provenance
**********

The Activity object represents the source of a data set or the data processing steps used to produce it. Using
`W3C provenance ontology <http://www.w3.org/2011/prov/wiki/Main_Page>`_ terms, a result is **generated by** a
combination of data and code which are either **used** or **executed**.

~~~~~~~
Imports
~~~~~~~

::

    from synapseclient import Activity

~~~~~~~~~~~~~~~~~~~~~~~~~~~
Creating an activity object
~~~~~~~~~~~~~~~~~~~~~~~~~~~

::

    act = Activity(name='clustering',
                   description='whizzy clustering',
                   used=['syn1234','syn1235'],
                   executed='syn4567')

Here, syn1234 and syn1235 might be two types of measurements on a common set of samples. Some whizzy clustering code
might be referred to by syn4567.  The used and executed can reference entities in Synapse or URLs.

Alternatively, you can build an activity up piecemeal::

    act = Activity(name='clustering', description='whizzy clustering')
    act.used(['syn12345', 'syn12346'])
    act.executed(
        'https://raw.githubusercontent.com/Sage-Bionetworks/synapsePythonClient/develop/tests/unit/unit_test_client.py')

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Storing entities with provenance
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The activity can be passed in when storing an Entity to set the Entity's provenance::

    clustered_samples = syn.store(clustered_samples, activity=act)

We've now recorded that `clustered_samples` is the output of our whizzy clustering algorithm applied to the data stored
in syn1234 and syn1235.

~~~~~~~~~~~~~~~~~~~~~
Recording data source
~~~~~~~~~~~~~~~~~~~~~

The :py:meth:`synapseclient.Synapse.store` has shortcuts for specifying the used and executed lists directly.
For example, when storing a data entity, it's a good idea to record its source::

    excellent_data = syn.store(excellent_data,
                               activityName='data-r-us'
                               activityDescription='downloaded from data-r-us',
                               used='http://data-r-us.com/excellent/data.xyz')


~~~~~~~~
Activity
~~~~~~~~

.. autoclass:: synapseclient.activity.Activity
   :members:

"""

import collections.abc

from synapseclient.core.exceptions import SynapseError, SynapseMalformedEntityError
from synapseclient.core.utils import is_url, is_synapse_id
from synapseclient.entity import is_synapse_entity


def is_used_entity(x):
    """Returns True if the given object represents a UsedEntity."""

    # A UsedEntity must be a dictionary with a 'reference' field, with a 'targetId' field
    if not isinstance(x, collections.abc.Mapping) \
            or 'reference' not in x \
            or 'targetId' not in x['reference']:
        return False

    # Must only have three keys
    if not all(key in ('reference', 'wasExecuted', 'concreteType') for key in x.keys()):
        return False

    # 'reference' field can only have two keys
    if not all(key in ('targetId', 'targetVersionNumber') for key in x['reference'].keys()):
        return False

    return True


def is_used_url(x):
    """Returns True if the given object represents a UsedURL."""

    # A UsedURL must be a dictionary with a 'url' field
    if not isinstance(x, collections.abc.Mapping) or 'url' not in x:
        return False

    # Must only have four keys
    if not all(key in ('url', 'name', 'wasExecuted', 'concreteType') for key in x.keys()):
        return False

    return True


def _get_any_bad_args(badargs, dictionary):
    """Returns the intersection of 'badargs' and the non-Null keys of 'dictionary'."""

    return list(illegal for illegal in badargs
                if illegal in dictionary and dictionary[illegal] is not None)


def _raise_incorrect_used_usage(badargs, message):
    """Raises an informative exception about Activity.used()."""

    if any(badargs):
        raise SynapseMalformedEntityError(
            "The parameter%s '%s' %s not allowed in combination with a %s." % (
                "s" if len(badargs) > 1 else "",
                badargs,
                "are" if len(badargs) > 1 else "is",
                message
            )
        )


class Activity(dict):
    """
    Represents the provenance of a Synapse Entity.

    :param name:        name of the Activity
    :param description: a short text description of the Activity
    :param used:        Either a list of:
                            - `reference objects \
                              <http://docs.synapse.org/rest/org/sagebionetworks/repo/model/Reference.html>`_
                              (e.g. ``[{'targetId':'syn123456', 'targetVersionNumber':1}]``)
                            - a list of Synapse Entities or Entity IDs
                            - a list of URL's
    :param executed:    A code resource that was executed to generate the Entity.
    :param data:        A dictionary representation of an Activity, with fields 'name', 'description' and 'used'
                        (a list of reference objects)

    See also: The `W3C's provenance ontology <http://www.w3.org/TR/prov-o/>`_
    """

    # TODO: make constructors from JSON consistent across objects
    def __init__(self, name=None, description=None, used=None, executed=None, data={}):
        super(Activity, self).__init__(data)
        if 'used' not in self:
            self['used'] = []

        if name is not None:
            self['name'] = name
        if description is not None:
            self['description'] = description
        if used is not None:
            self.used(used)
        if executed is not None:
            self.executed(executed)

    def used(self, target=None, targetVersion=None, wasExecuted=None, url=None, name=None):
        """
        Add a resource used by the activity.

        This method tries to be as permissive as possible. It accepts a string which might be a synapse ID or a URL,
        a synapse entity, a UsedEntity or UsedURL dictionary or a list containing any combination of these.

        In addition, named parameters can be used to specify the fields of either a UsedEntity or a UsedURL.
        If target and optionally targetVersion are specified, create a UsedEntity.
        If url and optionally name are specified, create a UsedURL.

        It is an error to specify both target/targetVersion parameters and url/name parameters in the same call.
        To add multiple UsedEntities and UsedURLs, make a separate call for each or pass in a list.

        In case of conflicting settings for wasExecuted both inside an object and with a parameter, the parameter wins.
        For example, this UsedURL will have wasExecuted set to False::

            activity.used({'url':'http://google.com', 'name':'Goog', 'wasExecuted':True}, wasExecuted=False)

        Entity examples::

            activity.used('syn12345')
            activity.used(entity)
            activity.used(target=entity, targetVersion=2)
            activity.used(codeEntity, wasExecuted=True)
            activity.used({'reference':{'target':'syn12345', 'targetVersion':1}, 'wasExecuted':False})

        URL examples::

            activity.used('http://mydomain.com/my/awesome/data.RData')
            activity.used(url='http://mydomain.com/my/awesome/data.RData', name='Awesome Data')
            activity.used(url='https://github.com/joe_hacker/code_repo', name='Gnarly hacks', wasExecuted=True)
            activity.used({'url':'https://github.com/joe_hacker/code_repo', 'name':'Gnarly hacks'}, wasExecuted=True)

        List example::

            activity.used(['syn12345', 'syn23456', entity, \
                          {'reference':{'target':'syn100009', 'targetVersion':2}, 'wasExecuted':True}, \
                          'http://mydomain.com/my/awesome/data.RData'])
        """
        # -- A list of targets
        if isinstance(target, list):
            badargs = _get_any_bad_args(['targetVersion', 'url', 'name'], locals())
            _raise_incorrect_used_usage(badargs, 'list of used resources')

            for item in target:
                self.used(item, wasExecuted=wasExecuted)
            return

        # -- UsedEntity
        elif is_used_entity(target):
            badargs = _get_any_bad_args(['targetVersion', 'url', 'name'], locals())
            _raise_incorrect_used_usage(badargs, 'dictionary representing a used resource')

            resource = target
            if 'concreteType' not in resource:
                resource['concreteType'] = 'org.sagebionetworks.repo.model.provenance.UsedEntity'

        # -- Used URL
        elif is_used_url(target):
            badargs = _get_any_bad_args(['targetVersion', 'url', 'name'], locals())
            _raise_incorrect_used_usage(badargs, 'URL')

            resource = target
            if 'concreteType' not in resource:
                resource['concreteType'] = 'org.sagebionetworks.repo.model.provenance.UsedURL'

        # -- Synapse Entity
        elif is_synapse_entity(target):
            badargs = _get_any_bad_args(['url', 'name'], locals())
            _raise_incorrect_used_usage(badargs, 'Synapse entity')

            reference = {'targetId': target['id']}
            if 'versionNumber' in target:
                reference['targetVersionNumber'] = target['versionNumber']
            if targetVersion:
                reference['targetVersionNumber'] = int(targetVersion)
            resource = {'reference': reference,
                        'concreteType': 'org.sagebionetworks.repo.model.provenance.UsedEntity'}
        # -- URL parameter
        elif url:
            badargs = _get_any_bad_args(['target', 'targetVersion'], locals())
            _raise_incorrect_used_usage(badargs, 'URL')

            resource = {'url': url, 'name': name if name else target,
                        'concreteType': 'org.sagebionetworks.repo.model.provenance.UsedURL'}

        # -- URL as a string
        elif is_url(target):
            badargs = _get_any_bad_args(['targetVersion'], locals())
            _raise_incorrect_used_usage(badargs, 'URL')
            resource = {'url': target, 'name': name if name else target,
                        'concreteType': 'org.sagebionetworks.repo.model.provenance.UsedURL'}

        # -- Synapse Entity ID (assuming the string is an ID)
        elif isinstance(target, str):
            badargs = _get_any_bad_args(['url', 'name'], locals())
            _raise_incorrect_used_usage(badargs, 'Synapse entity')
            vals = target.split('.')  # Handle synapseIds of from syn234.4
            if not is_synapse_id(vals[0]):
                raise ValueError('%s is not a valid Synapse id' % target)
            if len(vals) == 2:
                if targetVersion and int(targetVersion) != int(vals[1]):
                    raise ValueError('Two conflicting versions for %s were specified' % target)
                targetVersion = int(vals[1])
            reference = {'targetId': vals[0]}
            if targetVersion:
                reference['targetVersionNumber'] = int(targetVersion)
            resource = {'reference': reference, 'concreteType': 'org.sagebionetworks.repo.model.provenance.UsedEntity'}
        else:
            raise SynapseError('Unexpected parameters in call to Activity.used().')

        # Set wasExecuted
        if wasExecuted is None:
            # Default to False
            if 'wasExecuted' not in resource:
                resource['wasExecuted'] = False
        else:
            # wasExecuted parameter overrides setting in an object
            resource['wasExecuted'] = wasExecuted

        # Add the used resource to the activity
        self['used'].append(resource)

    def executed(self, target=None, targetVersion=None, url=None, name=None):
        """
        Add a code resource that was executed during the activity.
        See :py:func:`synapseclient.activity.Activity.used`
        """
        self.used(target=target, targetVersion=targetVersion, url=url, name=name, wasExecuted=True)

    def _getStringList(self, wasExecuted=True):
        usedList = []
        for source in [source for source in self['used'] if source.get('wasExecuted', False) == wasExecuted]:
            if source['concreteType'].endswith('UsedURL'):
                if source.get('name'):
                    usedList.append(source.get('name'))
                else:
                    usedList.append(source.get('url'))
            else:  # It is an entity for now
                tmpstr = source['reference']['targetId']
                if 'targetVersionNumber' in source['reference']:
                    tmpstr += '.%i' % source['reference']['targetVersionNumber']
                usedList.append(tmpstr)
        return usedList

    def _getExecutedStringList(self):
        return self._getStringList(wasExecuted=True)

    def _getUsedStringList(self):
        return self._getStringList(wasExecuted=False)

    def __str__(self):
        str = '%s\n  Executed:\n' % self.get('name', '')
        str += '\n'.join(self._getExecutedStringList())
        str += '  Used:\n'
        str += '\n'.join(self._getUsedStringList())
        return str



"""
Created on Sep 21, 2017

@author: bhoff

sleep while checking registered _listeners
"""
import time

_listeners = []


def add_listener(listener):
    if not callable(listener):
        raise ValueError("listener is not callable")
    _listeners.append(listener)


def clear_listeners():
    del _listeners[:]


def doze(secs, listener_check_interval_secs=0.1):
    end_time = time.time() + secs
    while time.time() < end_time:
        for listener in _listeners:
            listener()
        time.sleep(listener_check_interval_secs)



import collections
import collections.abc
import datetime
import importlib
import os
import threading
import platform

from synapseclient.core.utils import iso_to_datetime, snake_case

try:
    boto3 = importlib.import_module('boto3')
except ImportError:
    # boto is not a requirement to load this module,
    # we are able to optionally use functionality if it's available
    boto3 = None

STS_PERMISSIONS = set(['read_only', 'read_write'])

# default minimum life left on a cached token that we'll hand out.
DEFAULT_MIN_LIFE = datetime.timedelta(hours=1)


class _TokenCache(collections.OrderedDict):
    """A self pruning dictionary of STS tokens.
    It will prune itself as new keys are added, removing the oldest if the
    max_size is exceeded, and always removing all tokens that have expired
    on each additional insert."""

    def __init__(self, max_size):
        super().__init__()
        self.max_size = max_size

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._prune()

    def _prune(self):
        while len(self) > self.max_size:
            self.popitem(last=False)

        to_delete = []
        before_timestamp = datetime.datetime.utcnow().timestamp()
        for entity_id, token in self.items():
            expiration_iso_str = token['expiration']

            # our "iso_to_datetime" util naively assumes UTC ("Z") times which in practice STS tokens are
            if iso_to_datetime(expiration_iso_str).timestamp() < before_timestamp:
                to_delete.append(entity_id)
            else:
                break

        for entity_id in to_delete:
            del self[entity_id]


class StsTokenStore:
    """
    Cache STS tokens in memory for observed entity ids.
    An optimization for long lived Synapse objects that will interact with the same
    Synapse storage locations over and over again so they don't have to do a remote call
    to fetch a new token for every entity, which for e.g. small files can amount to
    non trivial overhead.
    """

    # each token is < 1k but given we don't know how long Python process will be running
    # (could be very long in a programmatic environment) we impose a limit on the maximum
    # number of tokens we will store in memory to prevent this optimization from becoming
    # a memory leak.
    DEFAULT_TOKEN_CACHE_SIZE = 5000

    def __init__(self, max_token_cache_size=DEFAULT_TOKEN_CACHE_SIZE):
        self._tokens = {p: _TokenCache(max_token_cache_size) for p in STS_PERMISSIONS}
        self._lock = threading.Lock()

    def get_token(self, syn, entity_id, permission, min_remaining_life: datetime.timedelta):
        with self._lock:
            utcnow = datetime.datetime.utcnow()
            token_cache = self._tokens.get(permission)
            if token_cache is None:
                raise ValueError(f"Invalid STS permission {permission}")

            token = token_cache.get(entity_id)
            if not token or (iso_to_datetime(token['expiration']) - utcnow) < min_remaining_life:
                # either there is no cached token or the remaining life on the token isn't enough so fetch new
                token = token_cache[entity_id] = self._fetch_token(syn, entity_id, permission)

        return token

    @staticmethod
    def _fetch_token(syn, entity_id, permission):
        return syn.restGET(f'/entity/{entity_id}/sts?permission={permission}')


EXPORT_TEMPLATE_STRINGS = {
    'bash': """\
export SYNAPSE_STS_S3_LOCATION="s3://{bucket}/{baseKey}"
export AWS_ACCESS_KEY_ID="{accessKeyId}"
export AWS_SECRET_ACCESS_KEY="{secretAccessKey}"
export AWS_SESSION_TOKEN="{sessionToken}"
""",
    'cmd': """\
set SYNAPSE_STS_S3_LOCATION="s3://{bucket}/{baseKey}"
set AWS_ACCESS_KEY_ID="{accessKeyId}"
set AWS_SECRET_ACCESS_KEY="{secretAccessKey}"
set AWS_SESSION_TOKEN="{sessionToken}"
""",
    'powershell': """\
$Env:SYNAPSE_STS_S3_LOCATION="s3://{bucket}/{baseKey}"
$Env:AWS_ACCESS_KEY_ID="{accessKeyId}"
$Env:AWS_SECRET_ACCESS_KEY="{secretAccessKey}"
$Env:AWS_SESSION_TOKEN="{sessionToken}"
"""
}


def _format_export_template_string(syn, entity_id, credentials, template_string):
    # as of http://sagebionetworks.jira.com/browse/PLFM-6226
    # the initial call to the back end should return the bucket info as well as the STS token.
    # in the event that it doesn't we make a separate call to get the info from the upload destination.
    # at some point we can probably remove this extra check/call, but depending on when the fix
    # for the above is deployed and when the initial STS python client is released, we include
    # the fall back to fetch the keys separately, but only if necessary.
    bucket_keys = {'bucket', 'baseKey'}
    if all(k in credentials for k in bucket_keys):
        subs = credentials
    else:
        upload_destination = syn._getDefaultUploadDestination(entity_id)
        subs = {**upload_destination, **credentials}

    # if for some reason we still don't have the bucket info, we just don't include
    # the path in the output
    if any(k not in subs for k in bucket_keys):
        template_string = template_string[template_string.find('\n') + 1:]

    return template_string.format(**subs)


def get_sts_credentials(syn, entity_id, permission, *, output_format='json', min_remaining_life=None):
    """See Synapse.get_sts_storage_token"""
    min_remaining_life = min_remaining_life or DEFAULT_MIN_LIFE

    value = syn._sts_token_store.get_token(syn, entity_id, permission, min_remaining_life)

    if output_format == 'boto':
        # the Synapse STS API returns camel cased keys that we need to convert to use with boto.
        # prefix with "aws_", convert to snake case, and exclude any other key/value pairs in the value
        # e.g. expiration
        return {"aws_{}".format(snake_case(k)): value[k] for k in (
            'accessKeyId', 'secretAccessKey', 'sessionToken'
        )}
    elif output_format == 'json':
        # pass through what server sent
        return value

    elif output_format == 'shell':
        # for "shell" we try to detect what is best for the system
        # assume bourne compatible output outside of windows
        if platform.system() == 'Windows' and 'bash' not in os.environ.get('SHELL', ''):
            if len(os.getenv('PSModulePath', '').split(os.pathsep)) >= 3:
                # https://stackoverflow.com/a/55598796
                output_format = 'powershell'
            else:
                output_format = 'cmd'
        else:
            output_format = 'bash'

    template_string = EXPORT_TEMPLATE_STRINGS.get(output_format)
    if not template_string:
        raise ValueError(f'Unrecognized output_format {output_format}')

    return _format_export_template_string(syn, entity_id, value, template_string)


def with_boto_sts_credentials(fn, syn, entity_id, permission):
    """A wrapper around a function that will get sts credentials and try to use them on the given
    function which should take a dictionary with the aws_access_key_id, aws_secret_access_key, and aws_session_token
    as keys. If the given function returns a boto error that looks like the token has expired
    it will retry once after fetching fresh credentials.

    The purpose is to be able to use potentially cached credentials in long running tasks while reducing
    worry that they will expire in the middle of running and cause an unrecoverable error.
    The alternative of fetching a fresh STS token for every request might be okay for a few large files
    but would greatly slow down transferring many small files.
    """

    for attempt in range(2):
        credentials = get_sts_credentials(syn, entity_id, permission, output_format='boto')
        try:
            response = fn(credentials)
        except boto3.exceptions.Boto3Error as ex:
            if 'ExpiredToken' in str(ex) and attempt == 0:
                continue
            else:
                raise

        return response


def is_boto_sts_transfer_enabled(syn):
    """
    Check if the boto/STS transfers are enabled in the Synapse configuration.
    If enabled then synapseclient will attempt to automatically use boto to upload
    and download from supported storage locations that are sts enabled.

    :param syn:         A Synapse client

    :returns: True if STS if enabled, False otherwise
    """
    return bool(boto3 and syn.use_boto_sts_transfers)


def is_storage_location_sts_enabled(syn, entity_id, location):
    """
    Returns whether the given storage location is enabled for STS.

    :param syn:         A Synapse client
    :param entity_id:   id of synapse entity whose storage location we want to check for sts access
    :param location:    a storage location id or an dictionary representing the location UploadDestination
                                these)
    :returns: True if STS if enabled for the location, False otherwise
    """
    if not location:
        return False

    if isinstance(location, collections.abc.Mapping):
        # looks like this is already an upload destination dict
        destination = location

    else:
        # otherwise treat it as a storage location id,
        destination = syn.restGET(
            f'/entity/{entity_id}/uploadDestination/{location}',
            endpoint=syn.fileHandleEndpoint
        )

    return destination.get('stsEnabled', False)



import os
import json

SESSION_CACHE_FILEPATH = os.path.expanduser("~/.synapseSession")


def get_most_recent_user():
    session_cache = _read_session_cache(SESSION_CACHE_FILEPATH)
    return session_cache.get("<mostRecent>")


def set_most_recent_user(username):
    cachedSessions = {"<mostRecent>": username}
    _write_session_cache(SESSION_CACHE_FILEPATH, cachedSessions)


def _read_session_cache(filepath):
    """Returns the JSON contents of CACHE_DIR/SESSION_FILENAME."""
    try:
        file = open(filepath, 'r')
        result = json.load(file)
        if isinstance(result, dict):
            return result
    except:  # noqa
        # If we cannot parse the cache for any reason, treat it as if the cache is empty
        pass
    return {}


def _write_session_cache(filepath, data):
    """Dumps the JSON data into CACHE_DIR/SESSION_FILENAME."""
    with open(filepath, 'w') as file:
        json.dump(data, file)
        file.write('\n')  # For compatibility with R's JSON parser



"""
******
Entity
******

The Entity class is the base class for all entities, including Project, Folder, File, and Link.

Entities are dictionary-like objects in which both object and dictionary notation (entity.foo or entity['foo']) can be
used interchangeably.

Imports::

    from synapseclient import Project, Folder, File, Link

.. autoclass:: synapseclient.entity.Entity

~~~~~~~
Project
~~~~~~~

.. autoclass:: synapseclient.entity.Project

~~~~~~
Folder
~~~~~~

.. autoclass:: synapseclient.entity.Folder

~~~~
File
~~~~

.. autoclass:: synapseclient.entity.File

Changing File Names
-------------------

A Synapse File Entity has a name separate from the name of the actual file it represents. When a file is uploaded to
Synapse, its filename is fixed, even though the name of the entity can be changed at any time. Synapse provides a way
to change this filename and the content-type of the file for future downloads by creating a new version of the file
with a modified copy of itself.  This can be done with the synapseutils.copy_functions.changeFileMetaData function.

>>> import synapseutils
>>> e = syn.get(synid)
>>> print(os.path.basename(e.path))  ## prints, e.g., "my_file.txt"
>>> e = synapseutils.changeFileMetaData(syn, e, "my_newname_file.txt")

Setting *fileNameOverride* will **not** change the name of a copy of the
file that's already downloaded into your local cache. Either rename the
local copy manually or remove it from the cache and re-download.:

>>> syn.cache.remove(e.dataFileHandleId)
>>> e = syn.get(e)
>>> print(os.path.basename(e.path))  ## prints "my_newname_file.txt"

~~~~
Link
~~~~

.. autoclass:: synapseclient.entity.Link

~~~~~~~~~~~~
Table Schema
~~~~~~~~~~~~

.. autoclass:: synapseclient.table.Schema


~~~~~~~~~~~~~~~~~~
Entity View Schema
~~~~~~~~~~~~~~~~~~

.. autoclass:: synapseclient.table.EntityViewSchema


~~~~~~~~~~~~~~~~
DockerRepository
~~~~~~~~~~~~~~~~

.. autoclass:: synapseclient.entity.DockerRepository

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Properties and annotations, implementation details
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

In Synapse, entities have both properties and annotations. Properties are used by the system, whereas annotations are
completely user defined. In the Python client, we try to present this situation as a normal object, with one set of
properties.

Printing an entity will show the division between properties and annotations.::

    print(entity)

Under the covers, an Entity object has two dictionaries, one for properties and one for annotations. These two
namespaces are distinct, so there is a possibility of collisions. It is recommended to avoid defining annotations with
names that collide with properties, but this is not enforced.::

    ## don't do this!
    entity.properties['description'] = 'One thing'
    entity.annotations['description'] = 'A different thing'

In case of conflict, properties will take precedence.::

    print(entity.description)
    #> One thing

Some additional ambiguity is entailed in the use of dot notation. Entity objects have their own internal properties
which are not persisted to Synapse. As in all Python objects, these properties are held in object.__dict__. For
example, this dictionary holds the keys 'properties' and 'annotations' whose values are both dictionaries themselves.

The rule, for either getting or setting is: first look in the object then look in properties, then look in annotations.
If the key is not found in any of these three, a get results in a ``KeyError`` and a set results in a new annotation
being created. Thus, the following results in a new annotation that will be persisted in Synapse::

    entity.foo = 'bar'

To create an object member variable, which will *not* be persisted in Synapse, this unfortunate notation is required::

    entity.__dict__['foo'] = 'bar'

As mentioned previously, name collisions are entirely possible.
Keys in the three namespaces can be referred to unambiguously like so::

    entity.__dict__['key']

    entity.properties.key
    entity.properties['key']

    entity.annotations.key
    entity.annotations['key']

Most of the time, users should be able to ignore these distinctions and treat Entities like normal Python objects.
End users should never need to manipulate items in __dict__.

See also:

- :py:mod:`synapseclient.annotations`

"""

import collections.abc
import itertools
import io
import os
import inspect
import urllib.parse as urllib_parse

from synapseclient.core import utils
from synapseclient.core.exceptions import SynapseMalformedEntityError
from synapseclient.core.models.dict_object import DictObject
from synapseclient.core.utils import id_of, itersubclasses


class Versionable(object):
    """An entity for which Synapse will store a version history."""

    _synapse_entity_type = 'org.sagebionetworks.repo.model.Versionable'
    _property_keys = ['versionNumber', 'versionLabel', 'versionComment', 'versionUrl', 'versions']


# TODO: inherit from UserDict.DictMixin?
#       http://docs.python.org/2/library/userdict.html#UserDict.DictMixin

# Alternate implementations include:
# - a naming convention to tag object members
# - keeping a list of 'transient' variables (the object members)
# - giving up on the dot notation (implemented in Entity2.py in commit e441fcf5a6963118bcf2b5286c67fc66c004f2b5 in the
# entity_object branch)
# - giving up on hiding the difference between properties and annotations

class Entity(collections.abc.MutableMapping):
    """
    A Synapse entity is an object that has metadata, access control, and potentially a file. It can represent data,
    source code, or a folder that contains other entities.

    Entities should typically be created using the constructors for specific subclasses such as Project, Folder or File.
    """

    _synapse_entity_type = 'org.sagebionetworks.repo.model.Entity'
    _property_keys = ['id', 'name', 'description', 'parentId',
                      'entityType', 'concreteType',
                      'uri', 'etag', 'annotations', 'accessControlList',
                      'createdOn', 'createdBy', 'modifiedOn', 'modifiedBy']
    _local_keys = []

    @classmethod
    def create(cls, properties=None, annotations=None, local_state=None):
        """
        Create an Entity or a subclass given dictionaries of properties and annotations, as might be received from the
        Synapse Repository.

        :param properties:  A map of Synapse properties

            If 'concreteType' is defined in properties, we create the proper subclass of Entity. If not, give back the
            type whose constructor was called:

            If passed an Entity as input, create a new Entity using the input entity as a prototype.

        :param annotations: A map of user defined annotations
        :param local_state: Internal use only
        """

        # Create a new Entity using an existing Entity as a prototype
        if isinstance(properties, Entity):
            if annotations is None:
                annotations = {}
            if local_state is None:
                local_state = {}
            annotations.update(properties.annotations)
            local_state.update(properties.local_state())
            properties = properties.properties
            if 'id' in properties:
                del properties['id']

        if cls == Entity \
                and 'concreteType' in properties \
                and properties['concreteType'] in entity_type_to_class:
            cls = entity_type_to_class[properties['concreteType']]
        return cls(properties=properties, annotations=annotations, local_state=local_state)

    @classmethod
    def getURI(cls, id):
        return '/entity/%s' % id

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)

        # Make really sure that properties and annotations exist before
        # any object methods get invoked. This is important because the
        # dot operator magic methods have been overridden and depend on
        # properties and annotations existing.
        obj.__dict__['properties'] = DictObject()
        obj.__dict__['annotations'] = DictObject()
        return obj

    def __init__(self, properties=None, annotations=None, local_state=None, parent=None, **kwargs):

        if properties:
            if isinstance(properties, collections.abc.Mapping):
                if 'annotations' in properties and isinstance(properties['annotations'], collections.abc.Mapping):
                    annotations.update(properties['annotations'])
                    del properties['annotations']
                self.__dict__['properties'].update(properties)
            else:
                raise SynapseMalformedEntityError(
                    'Unknown argument type: properties is a %s' % str(type(properties))
                )

        if annotations:
            if isinstance(annotations, collections.abc.Mapping):
                self.__dict__['annotations'].update(annotations)
            elif isinstance(annotations, str):
                self.properties['annotations'] = annotations
            else:
                raise SynapseMalformedEntityError(
                    'Unknown argument type: annotations is a %s' % str(type(annotations))
                )

        if local_state:
            if isinstance(local_state, collections.abc.Mapping):
                self.local_state(local_state)
            else:
                raise SynapseMalformedEntityError(
                    'Unknown argument type: local_state is a %s' % str(type(local_state))
                )

        for key in self.__class__._local_keys:
            if key not in self.__dict__:
                self.__dict__[key] = None

        # Extract parentId from parent
        if 'parentId' not in kwargs:
            if parent:
                try:
                    kwargs['parentId'] = id_of(parent)
                except Exception:
                    if isinstance(parent, Entity) and 'id' not in parent:
                        raise SynapseMalformedEntityError(
                            "Couldn't find 'id' of parent."
                            " Has it been stored in Synapse?"
                        )
                    else:
                        raise SynapseMalformedEntityError("Couldn't find 'id' of parent.")

        # Note: that this will work properly if derived classes declare their internal state variable *before* invoking
        # super(...).__init__(...)
        for key, value in kwargs.items():
            self.__setitem__(key, value)

        if 'concreteType' not in self:
            self['concreteType'] = self.__class__._synapse_entity_type

        # Only project can be top-level. All other entity types require parentId don't enforce this for generic Entity
        if 'parentId' not in self \
                and not isinstance(self, Project) \
                and not type(self) == Entity:
            raise SynapseMalformedEntityError("Entities of type %s must have a parentId." % type(self))

    def postURI(self):
        return '/entity'

    def putURI(self):
        return '/entity/%s' % self.id

    def deleteURI(self, versionNumber=None):
        if versionNumber:
            return '/entity/%s/version/%s' % (self.id, versionNumber)
        else:
            return '/entity/%s' % self.id

    def local_state(self, state=None):
        """
        Set or get the object's internal state, excluding properties, or annotations.

        :param state: A dictionary
        """
        if state:
            for key, value in state.items():
                if key not in ['annotations', 'properties']:
                    self.__dict__[key] = value
        result = {}
        for key, value in self.__dict__.items():
            if key not in ['annotations', 'properties'] and not key.startswith('__'):
                result[key] = value
        return result

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)

    def __setitem__(self, key, value):
        if key in self.__dict__ or key in self.__class__._local_keys:
            # If we assign like so:
            #   entity.annotations = {'foo';123, 'bar':'bat'}
            # Wrap the dictionary in a DictObject so we can
            # later do:
            #   entity.annotations.foo = 'bar'
            if (key == 'annotations' or key == 'properties') and not isinstance(value, DictObject):
                value = DictObject(value)
            self.__dict__[key] = value
        elif key in self.__class__._property_keys:
            self.properties[key] = value
        else:
            self.annotations[key] = value

    # TODO: def __delattr__

    def __getattr__(self, key):
        # Note: that __getattr__ is only called after an attempt to
        # look the key up in the object's dictionary has failed.
        try:
            return self.__getitem__(key)
        except KeyError:
            # Note that hasattr in Python2 is more permissive than Python3
            # about what exceptions it catches. In Python3, hasattr catches
            # only AttributeError
            raise AttributeError(key)

    def __getitem__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        elif key in self.properties:
            return self.properties[key]
        elif key in self.annotations:
            return self.annotations[key]
        else:
            raise KeyError(key)

    def __delitem__(self, key):
        if key in self.properties:
            del self.properties[key]
        elif key in self.annotations:
            del self.annotations[key]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    # TODO shouldn't these include local_state as well? -jcb
    def keys(self):
        """Returns a set of property and annotation keys"""
        return set(self.properties.keys()) | set(self.annotations.keys())

    def has_key(self, key):
        """Is the given key a property or annotation?"""

        return key in self.properties or key in self.annotations

    def _write_kvps(self, f, dictionary, key_filter=None, key_aliases=None):
        for key in sorted(dictionary.keys()):
            if (not key_filter) or key_filter(key):
                f.write('  ')
                f.write(str(key) if not key_aliases else key_aliases[key])
                f.write('=')
                f.write(str(dictionary[key]))
                f.write('\n')

    def __str__(self):
        f = io.StringIO()

        f.write('%s: %s (%s)\n' % (self.__class__.__name__, self.properties.get('name', 'None'),
                                   self['id'] if 'id' in self else '-',))

        self._str_localstate(f)

        f.write('properties:\n')
        self._write_kvps(f, self.properties)

        f.write('annotations:\n')
        self._write_kvps(f, self.annotations)

        return f.getvalue()

    def _str_localstate(self, f):  # type: (io.StringIO) -> None
        """
        Helper method for writing the string representation of the local state to a StringIO object
        :param f: a StringIO object to which the local state string will be written
        """
        self._write_kvps(f, self.__dict__,
                         lambda key: not (key in ['properties', 'annotations'] or key.startswith('__')))

    def __repr__(self):
        """Returns an eval-able representation of the Entity."""

        f = io.StringIO()
        f.write(self.__class__.__name__)
        f.write("(")
        f.write(", ".join(
            {"%s=%s" % (str(key), value.__repr__(),) for key, value in
                itertools.chain(
                    list([k_v for k_v in self.__dict__.items()
                          if not (k_v[0] in ['properties', 'annotations'] or k_v[0].startswith('__'))]),
                    self.properties.items(),
                    self.annotations.items())}))
        f.write(")")
        return f.getvalue()


class Project(Entity):
    """
    Represents a project in Synapse.

    Projects in Synapse must be uniquely named. Trying to create a project with a name that's already taken, say
    'My project', will result in an error

    :param name:            The name of the project
    :param properties:      A map of Synapse properties
    :param annotations:     A map of user defined annotations
    :param local_state:     Internal use only

    Example::

        project = Project('Foobarbat project')
        project = syn.store(project)
    """

    _synapse_entity_type = 'org.sagebionetworks.repo.model.Project'

    def __init__(self, name=None, properties=None, annotations=None, local_state=None, **kwargs):
        if name:
            kwargs['name'] = name
        super(Project, self).__init__(concreteType=Project._synapse_entity_type, properties=properties,
                                      annotations=annotations, local_state=local_state, **kwargs)


class Folder(Entity):
    """
    Represents a folder in Synapse.

    Folders must have a name and a parent and can optionally have annotations.

    :param name:            The name of the folder
    :param parent:          The parent project or folder
    :param properties:      A map of Synapse properties
    :param annotations:     A map of user defined annotations
    :param local_state:     Internal use only

    Example::

        folder = Folder('my data', parent=project)
        folder = syn.store(folder)
    """

    _synapse_entity_type = 'org.sagebionetworks.repo.model.Folder'

    def __init__(self, name=None, parent=None, properties=None, annotations=None, local_state=None, **kwargs):
        if name:
            kwargs['name'] = name
        super(Folder, self).__init__(concreteType=Folder._synapse_entity_type, properties=properties,
                                     annotations=annotations, local_state=local_state, parent=parent, **kwargs)


class Link(Entity):
    """
    Represents a link in Synapse.

    Links must have a target ID and a parent. When you do :py:func:`synapseclient.Synapse.get` on a Link object,
    the Link object is returned. If the target is desired, specify followLink=True in synapseclient.Synapse.get.

    :param targetId:        The ID of the entity to be linked
    :param targetVersion:   The version of the entity to be linked
    :param parent:          The parent project or folder
    :param properties:      A map of Synapse properties
    :param annotations:     A map of user defined annotations
    :param local_state:     Internal use only

    Example::

        link = Link('targetID', parent=folder)
        link = syn.store(link)
    """
    _property_keys = Entity._property_keys + ['linksTo', 'linksToClassName']
    _local_keys = Entity._local_keys
    _synapse_entity_type = 'org.sagebionetworks.repo.model.Link'

    def __init__(self, targetId=None, targetVersion=None, parent=None, properties=None, annotations=None,
                 local_state=None, **kwargs):
        if targetId is not None and targetVersion is not None:
            kwargs['linksTo'] = dict(targetId=utils.id_of(targetId), targetVersionNumber=targetVersion)
        elif targetId is not None and targetVersion is None:
            kwargs['linksTo'] = dict(targetId=utils.id_of(targetId))
        elif properties is not None and 'linksTo' in properties:
            pass
        else:
            raise SynapseMalformedEntityError("Must provide a target id")
        super(Link, self).__init__(concreteType=Link._synapse_entity_type, properties=properties,
                                   annotations=annotations, local_state=local_state, parent=parent, **kwargs)


class File(Entity, Versionable):
    """
    Represents a file in Synapse.

    When a File object is stored, the associated local file or its URL will be stored in Synapse. A File must have a
    path (or URL) and a parent.

    :param path:                Location to be represented by this File
    :param name:                Name of the file in Synapse, not to be confused with the name within the path
    :param parent:              Project or Folder where this File is stored
    :param synapseStore:        Whether the File should be uploaded or if only the path should be stored when
                                :py:func:`synapseclient.Synapse.store` is called on the File object.
                                Defaults to True (file should be uploaded)
    :param contentType:         Manually specify Content-type header, for example "application/png" or
                                "application/json; charset=UTF-8"
    :param dataFileHandleId:    Defining an existing dataFileHandleId will use the existing dataFileHandleId
                                The creator of the file must also be the owner of the dataFileHandleId to have
                                permission to store the file
    :param properties:          A map of Synapse properties
    :param annotations:         A map of user defined annotations
    :param local_state:         Internal use only

    Example::

        data = File('/path/to/file/data.xyz', parent=folder)
        data = syn.store(data)
    """
    # Note: externalURL technically should not be in the keys since it's only a field/member variable of
    # ExternalFileHandle, but for backwards compatibility it's included
    _file_handle_keys = ["createdOn", "id", "concreteType", "contentSize", "createdBy", "etag", "fileName",
                         "contentType", "contentMd5", "storageLocationId", 'externalURL']
    # Used for backwards compatability. The keys found below used to located in the entity's local_state
    # (i.e. __dict__).
    _file_handle_aliases = {'md5': 'contentMd5', 'externalURL': 'externalURL', 'fileSize': 'contentSize',
                            'contentType': 'contentType'}
    _file_handle_aliases_inverse = {v: k for k, v in _file_handle_aliases.items()}

    _property_keys = Entity._property_keys + Versionable._property_keys + ['dataFileHandleId']
    _local_keys = Entity._local_keys + ['path', 'cacheDir', 'files', 'synapseStore', '_file_handle']
    _synapse_entity_type = 'org.sagebionetworks.repo.model.FileEntity'

    # TODO: File(path="/path/to/file", synapseStore=True, parentId="syn101")
    def __init__(self, path=None, parent=None, synapseStore=True, properties=None,
                 annotations=None, local_state=None, **kwargs):
        if path and 'name' not in kwargs:
            kwargs['name'] = utils.guess_file_name(path)
        self.__dict__['path'] = path
        if path:
            cacheDir, basename = os.path.split(path)
            self.__dict__['cacheDir'] = cacheDir
            self.__dict__['files'] = [basename]
        else:
            self.__dict__['cacheDir'] = None
            self.__dict__['files'] = []
        self.__dict__['synapseStore'] = synapseStore

        # pop the _file_handle from local properties because it is handled differently from other local_state
        self._update_file_handle(local_state.pop('_file_handle', None) if (local_state is not None) else None)

        super(File, self).__init__(concreteType=File._synapse_entity_type, properties=properties,
                                   annotations=annotations, local_state=local_state, parent=parent, **kwargs)

    def _update_file_handle(self, file_handle_update_dict=None):
        """
        Sets the file handle

        Should not need to be called by users
        """

        # replace the file handle dict
        fh_dict = DictObject(file_handle_update_dict) if file_handle_update_dict is not None else DictObject()
        self.__dict__['_file_handle'] = fh_dict

        if file_handle_update_dict is not None \
                and file_handle_update_dict.get('concreteType') ==\
                "org.sagebionetworks.repo.model.file.ExternalFileHandle"\
                and urllib_parse.urlparse(file_handle_update_dict.get('externalURL')).scheme != 'sftp':
            self.__dict__['synapseStore'] = False

        # initialize all nonexistent keys to have value of None
        for key in self.__class__._file_handle_keys:
            if key not in fh_dict:
                fh_dict[key] = None

    def __setitem__(self, key, value):
        if key == '_file_handle':
            self._update_file_handle(value)
        elif key in self.__class__._file_handle_aliases:
            self._file_handle[self.__class__._file_handle_aliases[key]] = value
        else:
            def expand_and_convert_to_URL(path): return utils.as_url(os.path.expandvars(os.path.expanduser(path)))
            # hacky solution to allowing immediate switching into a ExternalFileHandle pointing to the current path
            # yes, there is boolean zen but I feel like it is easier to read/understand this way
            if key == 'synapseStore' and value is False and self['synapseStore'] is True \
                    and utils.caller_module_name(inspect.currentframe()) != 'client':
                self['externalURL'] = expand_and_convert_to_URL(self['path'])

            # hacky solution because we historically allowed modifying 'path' to indicate wanting to change to a new
            # ExternalFileHandle
            # don't change exernalURL if it's just the synapseclient setting metadata after a function call such as
            # syn.get()
            if key == 'path' and not self['synapseStore'] \
                    and utils.caller_module_name(inspect.currentframe()) != 'client':
                self['externalURL'] = expand_and_convert_to_URL(value)
                self['contentMd5'] = None
                self['contentSize'] = None
            super(File, self).__setitem__(key, value)

    def __getitem__(self, item):
        if item in self.__class__._file_handle_aliases:
            return self._file_handle[self.__class__._file_handle_aliases[item]]
        else:
            return super(File, self).__getitem__(item)

    def _str_localstate(self, f):
        self._write_kvps(f, self._file_handle, lambda key: key in ['externalURL', 'contentMd5', 'contentSize',
                                                                   'contentType'], self._file_handle_aliases_inverse)
        self._write_kvps(f, self.__dict__, lambda key: not (key in ['properties', 'annotations', '_file_handle']
                                                            or key.startswith('__')))


class DockerRepository(Entity):
    """
    A Docker repository is a lightweight virtual machine image.

    NOTE: store()-ing a DockerRepository created in the Python client will always result in it being treated as a
    reference to an external Docker repository that is not managed by synapse.
    To upload a docker image that is managed by Synapse please use the official Docker client and read
     http://docs.synapse.org/articles/docker.html for instructions on uploading a Docker Image to Synapse

    :param repositoryName: the name of the Docker Repository. Usually in the format: [host[:port]/]path.
     If host is not set, it will default to that of DockerHub. port can only be specified if the host is also specified.
    :param parent: the parent project for the Docker repository
    :param properties:      A map of Synapse properties
    :param annotations:     A map of user defined annotations
    :param local_state:     Internal use only

    :return:  an object of type :py:class:`synapseclient.entity.DockerRepository`
    """
    _synapse_entity_type = 'org.sagebionetworks.repo.model.docker.DockerRepository'

    _property_keys = Entity._property_keys + ['repositoryName']

    def __init__(self, repositoryName=None, parent=None, properties=None, annotations=None, local_state=None, **kwargs):
        if repositoryName:
            kwargs['repositoryName'] = repositoryName
        super(DockerRepository, self).__init__(properties=properties, annotations=annotations, local_state=local_state,
                                               parent=parent, **kwargs)
        if 'repositoryName' not in self:
            raise SynapseMalformedEntityError("DockerRepository must have a repositoryName.")


# Create a mapping from Synapse class (as a string) to the equivalent Python class.
entity_type_to_class = {}
for cls in itersubclasses(Entity):
    entity_type_to_class[cls._synapse_entity_type] = cls

_entity_types = ["project", "folder", "file", "table", "link", "entityview", "dockerrepo"]


def split_entity_namespaces(entity):
    """
    Given a plain dictionary or an Entity object, splits the object into properties, annotations and local state.
    A dictionary will be processed as a specific type of Entity if it has a valid 'concreteType' field, otherwise it is
    treated as a generic Entity.

    :returns: a 3-tuple (properties, annotations, local_state).
    """
    if isinstance(entity, Entity):
        # Defensive programming: return copies
        return entity.properties.copy(), entity.annotations.copy(), entity.local_state()

    if not isinstance(entity, collections.abc.Mapping):
        raise SynapseMalformedEntityError("Can't split a %s object." % entity.__class__.__name__)

    if 'concreteType' in entity and entity['concreteType'] in entity_type_to_class:
        entity_class = entity_type_to_class[entity['concreteType']]
    else:
        entity_class = Entity

    properties = DictObject()
    annotations = DictObject()
    local_state = DictObject()

    property_keys = entity_class._property_keys
    local_keys = entity_class._local_keys
    for key, value in entity.items():
        if key in property_keys:
            properties[key] = value
        elif key in local_keys:
            local_state[key] = value
        else:
            annotations[key] = value

    return properties, annotations, local_state


ENTITY_TYPES = [
    'org.sagebionetworks.repo.model.FileEntity',
    'org.sagebionetworks.repo.model.Folder',
    'org.sagebionetworks.repo.model.Link',
    'org.sagebionetworks.repo.model.Project',
    'org.sagebionetworks.repo.model.table.TableEntity'
]


def is_synapse_entity(entity):
    if isinstance(entity, Entity):
        return True
    if isinstance(entity, collections.abc.Mapping):
        return entity.get('concreteType', None) in ENTITY_TYPES
    return False


def is_versionable(entity):
    """Return True if the given entity's concreteType is one that is Versionable."""

    if isinstance(entity, Versionable):
        return True

    try:
        entity_class = entity_type_to_class[entity['concreteType']]
        return issubclass(entity_class, Versionable)
    except (KeyError, TypeError):
        # the dict input is not an entity
        raise ValueError("Input is not an entity.")


def is_container(entity):
    """Test if an entity is a container (ie, a Project or a Folder)"""
    if 'concreteType' in entity:
        concreteType = entity['concreteType']
    elif 'type' in entity:
        concreteType = entity['type']
    elif isinstance(entity, collections.abc.Mapping):
        prefix = utils.extract_prefix(entity.keys())
        if prefix+'concreteType' in entity:
            concreteType = entity[prefix+'concreteType'][0]
        elif prefix+'nodeType' in entity:
            return entity[prefix+'nodeType'] in ['project', 'folder']
        else:
            return False
    else:
        return False
    return concreteType in (Project._synapse_entity_type, Folder._synapse_entity_type)



"""
********
Overview
********

The ``synapseutils`` package provides both higher level functions as well as utilities for interacting with
`Synapse <http://www.synapse.org>`_.  These functionalities include:

- :py:func:`copy_functions.copy`
- :py:func:`copy_functions.copyWiki`
- :py:func:`walk.walk`
- :py:func:`sync.syncFromSynapse`
- :py:func:`sync.syncToSynapse`
- :py:func:`monitor.notifyMe`
- :py:func:`migrate_functions.index_files_for_migration`
- :py:func:`migrate_functions.migrate_indexed_files`
- :py:func:`describe_functions.describe`
"""
# flake8: noqa F401 unclear who is using these
from .copy_functions import copy, copyWiki, copyFileHandles, changeFileMetaData
from .walk import walk
from .sync import syncFromSynapse, syncToSynapse, generate_sync_manifest
from .migrate_functions import index_files_for_migration, migrate_indexed_files
from .monitor import notifyMe, with_progress_bar
from .describe_functions import describe



#
# Represent user-defined annotations on a synapse entity
# chris.bare@sagebase.org
############################################################

import collections.abc
import json


class DictObject(dict):

    @classmethod
    def getByNameURI(cls, name):
        print('%s can\'t be retrieved by name' % cls)
        raise ValueError

    def __init__(self, *args, **kwargs):
        self.__dict__ = self
        for arg in args:
            if isinstance(arg, collections.abc.Mapping):
                self.__dict__.update(arg)
        self.__dict__.update(kwargs)

    def __str__(self):
        return json.dumps(self, sort_keys=True, indent=2)

    def json(self, ensure_ascii=True):
        return json.dumps(self, sort_keys=True, indent=2, ensure_ascii=ensure_ascii)
